/*------------------------------------------------------------------------------
 lib/djv_base/math.h
 
 Math
 ----
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_MATH_H
#define DJV_MATH_H

#include <djv_base/base.h>

namespace djv {
namespace math
{

//------------------------------------------------------------------------------
// Constants

static const double pi = 3.1415927;
static const double pi_x2 = 6.2831853;
static const double pi_2 = 1.5707963;
static const double pi_4 = 0.7853981;

//------------------------------------------------------------------------------
// Functions
//
// References:
// * Andrew Glassner, "Graphics Gems".
// * David S. Ebert, F. Kenton Musgrave, Darwyn Peachey, Ken Perlin,
//   Steven Worley, "Texturing & Modeling, A Procedural Approach".

template<typename T>
inline T abs(T);

int    pow(int in, int value);
double pow(double in, double value);
double sqrt(double);
double exp(double);
double log(double);
double log_10(double);
double anti_log_10(double);

template<typename T>
inline T mod(T, T mod);
inline int wrap(int in, int min, int max);
inline int step(int in, int value);
inline int pulse(int in, int min, int max);

double bias(double in, double value);
double gain(double in, double value);

#if defined(min) // XXX WIN32.
#undef min
#endif
#if defined(max)
#undef max
#endif
template<typename T>
inline T min(T, T);
template<typename T>
inline T max(T, T);
template<typename T>
inline T clamp(T, T min, T max);

// Soft clip.
//
// References:
// * Ken McGaugh, "Re: [Openexr-devel] Cineon -> half conversion?", OpenEXR
//   Mailing List 3/2/05.

double soft_clip(double in, double value);

// Interpolation.

template<typename T, typename U>
inline T lerp(U in, T min, T max);

inline double smooth_step(double in, double min, double max);
  
// Trigonometry.

double sin(double);
double cos(double);
double sec(double);
double tan(double);

double arc_sin(double);
double arc_cos(double);
double arc_tan(double);
double arc_tan(double, double);

double co_sec(double);
double co_tan(double);

// Area.

template<typename T>
inline bool intersect(T, T min, T max);
template<typename T>
inline bool intersect(T min_a, T max_a, T min_b, T max_b, T * min, T * max);

template<typename T>
inline void bound(T min_a, T max_a, T min_b, T max_b, T * min, T * max);

// Random numbers.

double rand();
double rand(double);
double rand(double min, double max);

// Seed random value generator. A value of zero uses a random seed.

void rand_seed(uint = 0);

//------------------------------------------------------------------------------
// Conversion

//inline int ceil(double);
//inline int floor(double);
int ceil(double);
int floor(double);
inline double fraction(double);

inline int to_pow2(int);

inline double degrees_to_radians(double);
inline double radians_to_degrees(double);

void spherical_to_cartesian(double latitude, double longitude, double distance,
  double * x, double * y, double * z);
void cartesian_to_spherical(double x, double y, double z,
  double * latitude, double * longitude, double * distance);

}
}

#include <djv_base/math_.h>

#endif

