/*------------------------------------------------------------------------------
 lib/djv_image/gl_data.h
 
 OpenGL Image Data
 -----------------

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_GL_DATA_H
#define DJV_IMAGE_GL_DATA_H

#include <djv_base/box.h>
#include <djv_image/data.h>
#include <djv_image/color.h>
#include <djv_image/profile.h>
#include <djv_image/gl.h>

namespace djv {
namespace image {
namespace gl {
namespace data
{

//------------------------------------------------------------------------------
// Display Options

struct Color
{
  Color() : brightness(1), contrast(1), saturation(1) {}
  
  double brightness, contrast, saturation;
};
  
struct Levels
{
  Levels() : gamma(1) { in[0] = out[0] = 0; in[1] = out[1] = 1; }
  Levels(const Levels & _in) { *this = _in; }
  
  double in [2], gamma, out [2];
  
  Levels & operator = (const Levels &);
};

struct Display
{
  Display() : soft_clip(0.0) {}
  
  image::Data lut;
  Color color;
  Levels levels;
  double soft_clip;
};

//------------------------------------------------------------------------------
// Drawing Options

enum CHANNEL { CHANNEL_DEFAULT, CHANNEL_0, CHANNEL_1, CHANNEL_2, CHANNEL_3 };
enum FILTER { FILTER_LINEAR, FILTER_NEAREST };

struct Draw
{
  Draw() : mirror_x(false), mirror_y(false), rotate(0.0),
    channel(CHANNEL(0)), filter(FILTER(0))
  {}

  bool mirror_x, mirror_y;
  Vec2f position;
  Vec2f scale;
  double rotate;
  Profile profile;
  Display display;
  CHANNEL channel;
  FILTER filter;
  image::Color background;
};

//------------------------------------------------------------------------------
// Data

class Data
{
public:

  Data();
  ~Data();

  void draw(const image::Data &, const Box2i &, const Draw & = Draw(),
    void * context_share = 0) const throw (Error);
  
  static void read(image::Data *);
  static void read(image::Data *, const Box2i &);

private:

  class _Texture;
  class _Fixed;
  class _Shader;
  std::auto_ptr<_Texture> _texture;
  std::auto_ptr<_Fixed> _fixed;
  std::auto_ptr<_Shader> _shader;
  Draw _draw;
};

//------------------------------------------------------------------------------
// Functions

void copy(const image::Data &, image::Data *, const Draw & = Draw(),
  void * context_share = 0) throw (Error);

enum HISTOGRAM { HISTOGRAM_256, HISTOGRAM_1024 };
void histogram(const image::Data &, HISTOGRAM, image::Data *,
  image::Color * min, image::Color * max, const Draw & = Draw())
  throw (Error);
int histogram(HISTOGRAM);

void gradient(image::Data *) throw (Error);

Mtx4<GLfloat> color_mtx(const Color &);
image::Data color_lut(const Levels &, double soft_clip);

void unpack(const image::Info &, const Vec2i & offset = Vec2i());
void pack(const image::Info &, const Vec2i & offset = Vec2i());

//------------------------------------------------------------------------------
// Comparison

bool operator == (const Color &, const Color &);
bool operator != (const Color &, const Color &);

bool operator == (const Levels &, const Levels &);
bool operator != (const Levels &, const Levels &);

// XXX LUT data is not compared.

bool operator == (const Display &, const Display &);
bool operator != (const Display &, const Display &);

//------------------------------------------------------------------------------
// Conversion

bool serialize(List<Str> *, Color *);
List<Str> serialize(const Color &);

bool serialize(List<Str> *, Levels *);
List<Str> serialize(const Levels &);

// XXX LUT is not serialized.

bool serialize(List<Str> *, Display *);
List<Str> serialize(const Display &);

//------------------------------------------------------------------------------
// Labels

const List<Str> & label_channel();
const List<Str> & label_filter();
const List<Str> & label_histogram();

//------------------------------------------------------------------------------
// Global

void texture_tile(const Vec2i &);
const Vec2i & texture_tile();

void texture_dbg(bool);
bool texture_dbg();

//------------------------------------------------------------------------------
// Debugging

Dbg & operator << (Dbg &, const Color &);
Dbg & operator << (Dbg &, const Levels &);
Dbg & operator << (Dbg &, const Display &);
Dbg & operator << (Dbg &, CHANNEL);
Dbg & operator << (Dbg &, FILTER);

}
}
}
}

#endif

