/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_histogram.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/gl_data_histogram.cxx"

#include "image.h"

namespace djv {
namespace image {
namespace gl {
namespace data
{

namespace
{
int shift_10(HISTOGRAM in)
{
  switch (in) {
    case HISTOGRAM_256: return 2;
    case HISTOGRAM_1024: return 0;
  }
  return 0;
}

int shift_16(HISTOGRAM in)
{
  switch (in) {
    case HISTOGRAM_256: return 8;
    case HISTOGRAM_1024: return 6;
  }
  return 0;
}
}

void histogram(const image::Data & in, HISTOGRAM histogram,
  image::Data * out, image::Color * min, image::Color * max,
  const Draw & draw) throw (Error)
{
  DJV_ASSERT(out);

  DJV_DBG_OBJ2(_DBG, "histogram");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("histogram = " << data::histogram(histogram));
  
  const int size = data::histogram(histogram);
  out->set(image::data::Info(Vec2i(size, 1), pixel::RGB_U16));
  out->zero();
  DJV_DBG_PRN("out = " << *out);
  
  const PIXEL pixel = in.pixel();
  DJV_DBG_PRN("pixel = " << pixel);
  
  *min = image::Color(pixel);
  *max = image::Color(pixel);
  
  image::Data tmp(image::data::Info(in.size(), pixel));
  copy(in, &tmp, draw);
  
  const int w = tmp.w(), h = tmp.h();
  const int channels = tmp.channels();
  switch (pixel)
  {
    case pixel::RGB_U10:
    {
      const pixel::U10_S * in_p =
        reinterpret_cast<const pixel::U10_S *>(tmp.data());
      pixel::U16_T * const out_p =
        reinterpret_cast<pixel::U16_T *>(out->data());
      pixel::U10_S * const min_p =
        reinterpret_cast<pixel::U10_S *>(min->data());
      pixel::U10_S * const max_p =
        reinterpret_cast<pixel::U10_S *>(max->data());

      if (w && h)
      {
        min_p->r = max_p->r = in_p->r;
        min_p->g = max_p->g = in_p->g;
        min_p->b = max_p->b = in_p->b;
      }

      const int shift = shift_10(histogram);
      for (int y = 0; y < h; ++y)
        for (int x = 0; x < w; ++x, ++in_p)
        {
          out_p[(in_p->r >> shift) * 3 + 0]++;
          out_p[(in_p->g >> shift) * 3 + 1]++;
          out_p[(in_p->b >> shift) * 3 + 2]++;
          
          min_p->r = math::min(in_p->r, min_p->r);
          min_p->g = math::min(in_p->g, min_p->g);
          min_p->b = math::min(in_p->b, min_p->b);
          max_p->r = math::max(in_p->r, max_p->r);
          max_p->g = math::max(in_p->g, max_p->g);
          max_p->b = math::max(in_p->b, max_p->b);
        }
    } break;

#define _WORK(TYPE) \
      \
      const pixel::TYPE##_T * in_p = \
        reinterpret_cast<const pixel::TYPE##_T *>(tmp.data()); \
      pixel::U16_T * const out_p = \
        reinterpret_cast<pixel::U16_T *>(out->data()); \
      pixel::TYPE##_T * const min_p = \
        reinterpret_cast<pixel::TYPE##_T *>(min->data()); \
      pixel::TYPE##_T * const max_p = \
        reinterpret_cast<pixel::TYPE##_T *>(max->data()); \
      \
      if (w && h) \
        for (int c = 0; c < channels; ++c) \
          min_p[c] = max_p[c] = in_p[c]; \
      \
      const int shift = shift_16(histogram); \
      for (int y = 0; y < h; ++y) \
        for (int x = 0; x < w; ++x, in_p += channels) \
        { \
          for (int c = 0; c < channels && c < 3; ++c) \
            out_p[(PIXEL_##TYPE##_TO_U16(in_p[c]) >> shift) * 3 + c]++; \
          for (int c = 0; c < channels; ++c) { \
            min_p[c] = math::min(in_p[c], min_p[c]); \
            max_p[c] = math::max(in_p[c], max_p[c]); \
          } \
        }

    case pixel::L_U8:
    case pixel::LA_U8:
    case pixel::RGB_U8:
    case pixel::RGBA_U8: { _WORK(U8); } break;
    case pixel::L_U16:
    case pixel::LA_U16:
    case pixel::RGB_U16:
    case pixel::RGBA_U16: { _WORK(U16); } break;
    case pixel::L_F16:
    case pixel::LA_F16:
    case pixel::RGB_F16:
    case pixel::RGBA_F16: { _WORK(F16); } break;
    case pixel::L_F32:
    case pixel::LA_F32:
    case pixel::RGB_F32:
    case pixel::RGBA_F32: { _WORK(F32); } break;
  }
  
/*#if defined(DJV_DBG)
  pixel::U16_T * const out_p =
    reinterpret_cast<pixel::U16_T *>(out->data());
  for (int i = 0; i < size; ++i)
    DJV_DBG_PRN(i << " = " << out_p[i * 3]);
#endif*/
}

int histogram(HISTOGRAM in)
{
  static const int data [] = { 256, 1024 };
  return data[in];
}

const List<Str> & label_histogram()
{
  static const List<Str> data = List<Str>() << "256" << "1024";
  return data;
}

}
}
}
}

