/*------------------------------------------------------------------------------
 lib/djv_image/io_dpx_header.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_dpx_header.cxx"

#include "io_dpx.h"

namespace djv {
namespace image {
namespace io {
namespace dpx
{

namespace
{
enum ORIENT
{
  ORIENT_LEFT_RIGHT_TOP_BOTTOM,
  ORIENT_RIGHT_LEFT_TOP_BOTTOM,
  ORIENT_LEFT_RIGHT_BOTTOM_TOP,
  ORIENT_RIGHT_LEFT_BOTTOM_TOP,
  ORIENT_TOP_BOTTOM_LEFT_RIGHT,
  ORIENT_TOP_BOTTOM_RIGHT_LEFT,
  ORIENT_BOTTOM_TOP_LEFT_RIGHT,
  ORIENT_BOTTOM_TOP_RIGHT_LEFT
};

enum DESCRIPTOR
{
  DESCRIPTOR_USER = 0,
  DESCRIPTOR_R = 1,
  DESCRIPTOR_G = 2,
  DESCRIPTOR_B = 3,
  DESCRIPTOR_A = 4,
  DESCRIPTOR_L = 6,
  DESCRIPTOR_CHROMA = 7,
  DESCRIPTOR_Z = 8,
  DESCRIPTOR_COMPOSITE_VIDEO = 9,
  DESCRIPTOR_RGB = 50,
  DESCRIPTOR_RGBA = 51,
  DESCRIPTOR_ABGR = 52,
  DESCRIPTOR_CBYCRY = 100,
  DESCRIPTOR_CBYACRYA = 101,
  DESCRIPTOR_CBYCR = 102,
  DESCRIPTOR_CBYCRA = 103,
  DESCRIPTOR_USER_2 = 150,
  DESCRIPTOR_USER_3 = 151,
  DESCRIPTOR_USER_4 = 152,
  DESCRIPTOR_USER_5 = 153,
  DESCRIPTOR_USER_6 = 154,
  DESCRIPTOR_USER_7 = 155,
  DESCRIPTOR_USER_8 = 156
};

enum TRANSFER
{
  TRANSFER_USER,
  TRANSFER_PRINT,
  TRANSFER_LINEAR,
  TRANSFER_LOG,
  TRANSFER_VIDEO,
  TRANSFER_SMPTE_274M,
  TRANSFER_ITU_R_709_4,
  TRANSFER_ITU_R_601_5_B_OR_G,
  TRANSFER_ITU_R_601_5_M,
  TRANSFER_NTSC,
  TRANSFER_PAL,
  TRANSFER_Z,
  TRANSFER_Z_HOMOGENEOUS
};

enum COLORIMETRIC
{
  COLORIMETRIC_USER = 0,
  COLORIMETRIC_PRINT = 1,
  COLORIMETRIC_VIDEO = 4,
  COLORIMETRIC_SMPTE_274M = 5,
  COLORIMETRIC_ITU_R_709_4 = 6,
  COLORIMETRIC_ITU_R_601_5_B_OR_G = 7,
  COLORIMETRIC_ITU_R_601_5_M = 8,
  COLORIMETRIC_NTSC = 9,
  COLORIMETRIC_PAL = 10
};

enum COMPONENTS { PACK, TYPE_A, TYPE_B };

static const char magic [][5] = { "SDPX", "XPDS" };

struct Header
{
  struct File
  {
    uint32_t magic;
    uint32_t image_offset;
    char version [8];
    uint32_t size;
    uint32_t ditto_key;
    uint32_t header_size;
    uint32_t industry_header_size;
    uint32_t user_header_size;
    char name [100];
    char time [24];
    char creator [100];
    char project [200];
    char copyright [200];
    uint32_t encryption_key;
    uint8_t _pad [104];
  } file;
  
  struct Image
  {
    uint16_t orient;
    uint16_t elem_size;
    uint32_t size [2];
    
    struct Elem
    {
      uint32_t data_sign;
      uint32_t low_data;
      float low_quantity;
      uint32_t high_data;
      float high_quantity;
      uint8_t descriptor;
      uint8_t transfer;
      uint8_t colorimetric;
      uint8_t bit_depth;
      uint16_t packing, encoding;
      uint32_t data_offset;
      uint32_t line_padding;
      uint32_t elem_padding;
      char description [32];
    } elem [8];

    uint8_t _pad [52];
  } image;

  struct Source
  {
    uint32_t offset [2];
    float center [2];
    uint32_t size [2];
    char file [100];
    char time [24];
    char input_device [32];
    char input_serial [32];
    uint16_t border [4];
    uint32_t pixel_aspect [2];
    float scan_size [2];
    uint8_t _pad [20];
  } source;

  struct Film
  {
    char id [2];
    char type [2];
    char offset [2];
    char prefix [6];
    char count [4];
    char format [32];
    uint32_t frame;
    uint32_t sequence;
    uint32_t hold;
    float frame_rate;
    float shutter;
    char frame_id [32];
    char slate [100];
    uint8_t _pad [56];
  } film;

  struct Tv
  {
    uint32_t timecode;
    uint32_t user_bits;
    uint8_t interlace;
    uint8_t field, video_signal;
    uint8_t _pad;
    float sample_rate [2];
    float frame_rate;
    float time_offset;
    float gamma;
    float black_level;
    float black_gain;
    float breakpoint;
    float white_level;
    float integration_times;
    uint8_t _pad2 [76];
  } tv;
};

inline void zero(char * in, int size) { mem::zero(in, size); }

bool valid(const uint8_t * in) { return *in != 0xff; }
bool valid(const uint16_t * in) { return *in != 0xffff; }

// XXX Catch uninitialized values.

const int32_t _int_max = 1000000;
const float _float_max = 1000000.0;
const char _char_min = 32;
const char _char_max = 126;

const float _speed_min = 0.000001;

bool valid(const uint32_t * in)
{ return *in != 0xffffffff && *in < uint32_t(_int_max); }
bool valid(const float * in)
{ return *((uint32_t *)in) != 0xffffffff &&
  *in > -_float_max && *in < _float_max; }

bool valid(const char * in, int size)
{
  const char * p = in;
  const char * const end = p + size;
  for (; *p && p < end; ++p)
    if (*p < _char_min || *p > _char_max)
      return false;
  return size ? (in[0] != 0) : false;
}

inline Str _str(const char * in, int size)
{
  const char * p = in;
  const char * const end = p + size;
  for (; *p && p < end; ++p) ;
  return Str(in, p - in);
}

void _endian(Header * in)
{
  mem::endian(&in->file.image_offset, 1, 4);
  mem::endian(&in->file.size, 1, 4);
  mem::endian(&in->file.ditto_key, 1, 4);
  mem::endian(&in->file.header_size, 1, 4);
  mem::endian(&in->file.industry_header_size, 1, 4);
  mem::endian(&in->file.user_header_size, 1, 4);
  mem::endian(&in->file.encryption_key, 1, 4);
  
  mem::endian(&in->image.orient, 1, 2);
  mem::endian(&in->image.elem_size, 1, 2);
  mem::endian(in->image.size, 2, 4);
  for (uint i = 0; i < 8; ++i)
  {
    mem::endian(&in->image.elem[i].data_sign, 1, 4);
    mem::endian(&in->image.elem[i].low_data, 1, 4);
    mem::endian(&in->image.elem[i].low_quantity, 1, 4);
    mem::endian(&in->image.elem[i].high_data, 1, 4);
    mem::endian(&in->image.elem[i].high_quantity, 1, 4);
    mem::endian(&in->image.elem[i].packing, 1, 2);
    mem::endian(&in->image.elem[i].encoding, 1, 2);
    mem::endian(&in->image.elem[i].data_offset, 1, 4);
    mem::endian(&in->image.elem[i].line_padding, 1, 4);
    mem::endian(&in->image.elem[i].elem_padding, 1, 4);
  }
  
  mem::endian(in->source.offset, 2, 4);
  mem::endian(in->source.center, 2, 4);
  mem::endian(in->source.size, 2, 4);
  mem::endian(in->source.border, 4, 2);
  mem::endian(in->source.pixel_aspect, 2, 4);
  mem::endian(in->source.scan_size, 2, 4);
  
  mem::endian(&in->film.frame, 1, 4);
  mem::endian(&in->film.sequence, 1, 4);
  mem::endian(&in->film.hold, 1, 4);
  mem::endian(&in->film.frame_rate, 1, 4);
  mem::endian(&in->film.shutter, 1, 4);

  mem::endian(&in->tv.timecode, 1, 4);
  mem::endian(&in->tv.user_bits, 1, 4);
  mem::endian(in->tv.sample_rate, 2, 4);
  mem::endian(&in->tv.frame_rate, 1, 4);
  mem::endian(&in->tv.time_offset, 1, 4);
  mem::endian(&in->tv.gamma, 1, 4);
  mem::endian(&in->tv.black_level, 1, 4);
  mem::endian(&in->tv.black_gain, 1, 4);
  mem::endian(&in->tv.breakpoint, 1, 4);
  mem::endian(&in->tv.white_level, 1, 4);
  mem::endian(&in->tv.integration_times, 1, 4);
}

Str dbg(uint8_t in)
//{ return str::int_to_str(in); }
{ return valid(&in) ? str::int_to_str(in) : "-"; }
Str dbg(uint16_t in)
//{ return str::int_to_str(in); }
{ return valid(&in) ? str::int_to_str(in) : "-"; }
Str dbg(uint32_t in)
//{ return str::int_to_str(in); }
{ return valid(&in) ? str::int_to_str(in) : "-"; }
//{ return dbg_u32(in); }
Str dbg(float in)
//{ return str::float_to_str(in); }
{ return valid(&in) ? str::float_to_str(in) : "-"; }
//{ return dbg_u32(in); }
Str dbg(const char * in, int size)
//{ return Str(in, size); }
{ return valid(in, size) ? Str(in, size) : "-"; }

void dbg(const Header & in)
{
  DJV_DBG_OBJ2(_DBG, "dbg");
  
  DJV_DBG_PRN("file");
  DJV_DBG_PRN("  image offset = " << dbg(in.file.image_offset));
  DJV_DBG_PRN("  version = " << dbg(in.file.version, 8));
  DJV_DBG_PRN("  file size = " << dbg(in.file.size));
  DJV_DBG_PRN("  ditto key = " << dbg(in.file.ditto_key));
  DJV_DBG_PRN("  header size = " << dbg(in.file.header_size));
  DJV_DBG_PRN("  industry header size = " << dbg(in.file.industry_header_size));
  DJV_DBG_PRN("  user header size = " << dbg(in.file.user_header_size));
  DJV_DBG_PRN("  name = " << dbg(in.file.name, 100));
  DJV_DBG_PRN("  time = " << dbg(in.file.time, 24));
  DJV_DBG_PRN("  creator = " << dbg(in.file.creator, 100));
  DJV_DBG_PRN("  project = " << dbg(in.file.project, 200));
  DJV_DBG_PRN("  copyright = " << dbg(in.file.copyright, 200));
  DJV_DBG_PRN("  encryption key = " << dbg(in.file.encryption_key));

  DJV_DBG_PRN("image");
  DJV_DBG_PRN("  orient = " << dbg(in.image.orient));
  DJV_DBG_PRN("  element size = " << dbg(in.image.elem_size));
  DJV_DBG_PRN("  size = " << dbg(in.image.size[0]) << " " <<
    dbg(in.image.size[1]));
  for (int i = 0; i < in.image.elem_size; ++i)
  {
    DJV_DBG_PRN("  element " << i);
    DJV_DBG_PRN("    data sign = " << dbg(in.image.elem[i].data_sign));
    DJV_DBG_PRN("    low data = " << dbg(in.image.elem[i].low_data));
    DJV_DBG_PRN("    low quantity = " << dbg(in.image.elem[i].low_quantity));
    DJV_DBG_PRN("    high data = " << dbg(in.image.elem[i].high_data));
    DJV_DBG_PRN("    high quantity = " << dbg(in.image.elem[i].high_quantity));
    DJV_DBG_PRN("    descriptor = " << dbg(in.image.elem[i].descriptor));
    DJV_DBG_PRN("    transfer = " << dbg(in.image.elem[i].transfer));
    DJV_DBG_PRN("    colorimetric = " << dbg(in.image.elem[i].colorimetric));
    DJV_DBG_PRN("    bit depth = " << dbg(in.image.elem[i].bit_depth));
    DJV_DBG_PRN("    packing = " << dbg(in.image.elem[i].packing));
    DJV_DBG_PRN("    encoding = " << dbg(in.image.elem[i].encoding));
    DJV_DBG_PRN("    data offset = " << dbg(in.image.elem[i].data_offset));
    DJV_DBG_PRN("    line padding = " << dbg(in.image.elem[i].line_padding));
    DJV_DBG_PRN("    elem padding = " << dbg(in.image.elem[i].elem_padding));
    DJV_DBG_PRN("    description = " << dbg(in.image.elem[i].description, 32));
  }
  
  DJV_DBG_PRN("source");
  DJV_DBG_PRN("  offset = " << dbg(in.source.offset[0]) << " " <<
    dbg(in.source.offset[1]));
  DJV_DBG_PRN("  center = " << dbg(in.source.center[0]) << " " <<
    dbg(in.source.center[1]));
  DJV_DBG_PRN("  size = " << dbg(in.source.size[0]) << " " <<
    dbg(in.source.size[1]));
  DJV_DBG_PRN("  file = " << dbg(in.source.file, 100));
  DJV_DBG_PRN("  time = " << dbg(in.source.time, 24));
  DJV_DBG_PRN("  input device = " << dbg(in.source.input_device, 32));
  DJV_DBG_PRN("  inputs serial = " << dbg(in.source.input_serial, 32));
  DJV_DBG_PRN("  border = " << dbg(in.source.border[0]) << " " <<
    dbg(in.source.border[1]) << " " <<  dbg(in.source.border[2]) << " " <<
    dbg(in.source.border[3]));
  DJV_DBG_PRN("  pixel aspect = " << dbg(in.source.pixel_aspect[0]) << " " <<
    dbg(in.source.pixel_aspect[1]));
  DJV_DBG_PRN("  scan size = " << dbg(in.source.scan_size[0]) << " " <<
    dbg(in.source.scan_size[1]));

  DJV_DBG_PRN("film");
  DJV_DBG_PRN("  id = " << dbg(in.film.id, 2));
  DJV_DBG_PRN("  type = " << dbg(in.film.type, 2));
  DJV_DBG_PRN("  offset = " << dbg(in.film.offset, 2));
  DJV_DBG_PRN("  prefix = " << dbg(in.film.prefix, 6));
  DJV_DBG_PRN("  count = " << dbg(in.film.count, 4));
  DJV_DBG_PRN("  format = " << dbg(in.film.format, 32));
  DJV_DBG_PRN("  frame = " << dbg(in.film.frame));
  DJV_DBG_PRN("  sequence = " << dbg(in.film.sequence));
  DJV_DBG_PRN("  hold = " << dbg(in.film.hold));
  DJV_DBG_PRN("  frame rate = " << dbg(in.film.frame_rate));
  DJV_DBG_PRN("  shutter = " << dbg(in.film.shutter));
  DJV_DBG_PRN("  frame id = " << dbg(in.film.frame_id, 32));
  DJV_DBG_PRN("  slate = " << dbg(in.film.slate, 100));

  DJV_DBG_PRN("tv");
  DJV_DBG_PRN("  timecode = " << dbg(in.tv.timecode));
  DJV_DBG_PRN("  user bits = " << dbg(in.tv.user_bits));
  DJV_DBG_PRN("  interlace = " << dbg(in.tv.interlace));
  DJV_DBG_PRN("  field = " << dbg(in.tv.field));
  DJV_DBG_PRN("  video signal = " << dbg(in.tv.video_signal));
  DJV_DBG_PRN("  sample rate = " << dbg(in.tv.sample_rate[0]) << " " <<
    dbg(in.tv.sample_rate[1]));
  DJV_DBG_PRN("  frame rate = " << dbg(in.tv.frame_rate));
  DJV_DBG_PRN("  time offset = " << dbg(in.tv.time_offset));
  DJV_DBG_PRN("  gamma = " << dbg(in.tv.gamma));
  DJV_DBG_PRN("  black level = " << dbg(in.tv.black_level));
  DJV_DBG_PRN("  black gain = " << dbg(in.tv.black_gain));
  DJV_DBG_PRN("  breakpoint = " << dbg(in.tv.breakpoint));
  DJV_DBG_PRN("  white level = " << dbg(in.tv.white_level));
  DJV_DBG_PRN("  integration times = " << dbg(in.tv.integration_times));
}

void _set(const Info & info, COLOR color, Header::File * file,
  Header::Image * image)
{
  zero(file->version, 8);
  mem::copy("V2.0", file->version, 4);
  
  file->image_offset = 2048;
  file->header_size = 2048 - 384;
  file->industry_header_size = 384;
  file->user_header_size = 0;
  file->size = 0;
  file->ditto_key = 1;
  
  image->elem_size = 1;
  image->size[0] = info.size.x;
  image->size[1] = info.size.y;
  image->orient = ORIENT_LEFT_RIGHT_TOP_BOTTOM;
  
  switch (info.pixel)
  {
    case pixel::L_U8:
    case pixel::L_U16:
    case pixel::L_F16: 
    case pixel::L_F32: image->elem[0].descriptor = DESCRIPTOR_L; break;
    case pixel::RGB_U8:
    case pixel::RGB_U10:
    case pixel::RGB_U16:
    case pixel::RGB_F16: 
    case pixel::RGB_F32: image->elem[0].descriptor = DESCRIPTOR_RGB; break;
    case pixel::RGBA_U8:
    case pixel::RGBA_U16:
    case pixel::RGBA_F16: 
    case pixel::RGBA_F32: image->elem[0].descriptor = DESCRIPTOR_RGBA; break;
    default: break;
  }
  
  switch (info.pixel)
  {
    case pixel::RGB_U10: image->elem[0].packing = TYPE_A; break;
    default: break;
  }
  
  const int bit_depth = pixel::bit_depth(info.pixel);  
  image->elem[0].bit_depth = bit_depth;

  image->elem[0].data_sign = 0;
  image->elem[0].low_data = 0;
  switch (bit_depth)
  {
    case 8: image->elem[0].high_data = 255; break;
    case 10: image->elem[0].high_data = 1023; break;
    case 12: image->elem[0].high_data = 4095; break;
    case 16: image->elem[0].high_data = 65535; break;
  }
    
  if (COLOR_RAW == color)
  {
    image->elem[0].transfer = TRANSFER_LINEAR;
    image->elem[0].colorimetric = COLORIMETRIC_USER;
  }
  else
  {
    image->elem[0].transfer = TRANSFER_PRINT;
    image->elem[0].colorimetric = COLORIMETRIC_PRINT;
  }

  image->elem[0].encoding = 0;
  image->elem[0].data_offset = 2048;
  image->elem[0].line_padding = 0;
  image->elem[0].elem_padding = 0;
}

void _get(const Header::Image & in, Info * info, bool * print) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_get");

  if (in.elem_size != 1)
    throw Error(name, error_unsupported);
  
  info->size = Vec2i(in.size[0], in.size[1]);
  DJV_DBG_PRN("size = " << info->size);
  
  switch (in.orient)
  {
    case ORIENT_LEFT_RIGHT_TOP_BOTTOM: info->mirror_y = true; break;
    case ORIENT_RIGHT_LEFT_TOP_BOTTOM: info->mirror_x = true; break;
    case ORIENT_LEFT_RIGHT_BOTTOM_TOP: break;
    case ORIENT_RIGHT_LEFT_BOTTOM_TOP:
      info->mirror_x = info->mirror_y = true; break;
    case ORIENT_TOP_BOTTOM_LEFT_RIGHT:
    case ORIENT_TOP_BOTTOM_RIGHT_LEFT:
    case ORIENT_BOTTOM_TOP_LEFT_RIGHT:
    case ORIENT_BOTTOM_TOP_RIGHT_LEFT: break;
  }

  int pixel = -1;
  switch (in.elem[0].packing)
  {
    case PACK:
    {
      int channels = 0;
      switch (in.elem[0].descriptor)
      {
        case DESCRIPTOR_L: channels = 1; break;
        case DESCRIPTOR_RGB: channels = 3; break;
        case DESCRIPTOR_RGBA: channels = 4; break;
      }
      pixel = pixel::pixel(channels, in.elem[0].bit_depth);
    } break;
    case TYPE_A:
      switch (in.elem[0].descriptor)
      {
        case DESCRIPTOR_RGB:
          switch (in.elem[0].bit_depth)
          {
            case 10: pixel = pixel::RGB_U10; break;
          }
          break;
      }
      break;
  }
  if (-1 == pixel)
    throw Error(name, error_unsupported);
  info->pixel = PIXEL(pixel);
  DJV_DBG_PRN("pixel = " << info->pixel);
  
  switch (pixel::bit_depth(info->pixel))
  {
    case 8:
    case 10: info->align = 4; break;
  }

  if (in.elem[0].encoding)
    throw Error(name, error_unsupported);
  
  if (valid(&in.elem[0].line_padding) && in.elem[0].line_padding)
    throw Error(name, error_unsupported);
  
  if (print)
    *print = TRANSFER_PRINT == in.elem[0].transfer;
}

void _tag(const Str & file_name, const Tag & tag, Header::File * file,
  Header::Source * source, Header::Film * film, Header::Tv * tv)
{
  const List<Str> & label = tag::label();
  const List<Str> & dpx_label = dpx::label_tag();
  Str tmp;
  
  // File.
  
  zero(file->name, 100);
  str::ascii(file_name, file->name, 100, false);

  zero(file->time, 24);
  tmp = tag[label[tag::TIME]];
  if (tmp.size()) str::ascii(tmp, file->time, 24, false);

  zero(file->creator, 100);
  tmp = tag[label[tag::CREATOR]];
  if (tmp.size()) str::ascii(tmp, file->creator, 100, false);

  zero(file->project, 200);
  tmp = tag[label[tag::PROJECT]];
  if (tmp.size()) str::ascii(tmp, file->project, 200, false);

  zero(file->copyright, 200);
  tmp = tag[label[tag::COPYRIGHT]];
  if (tmp.size()) str::ascii(tmp, file->copyright, 200, false);
  
  file->encryption_key = 0;
  
  // Source.
  
  tmp = tag[dpx_label[TAG_SOURCE_OFFSET]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, " ");
    if (2 == list.size())
    {
      source->offset[0] = str::str_to_int(list[0]);
      source->offset[1] = str::str_to_int(list[1]);
    }
  }

  tmp = tag[dpx_label[TAG_SOURCE_CENTER]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, " ");
    if (2 == list.size())
    {
      source->center[0] = str::str_to_float(list[0]);
      source->center[1] = str::str_to_float(list[1]);
    }
  }

  tmp = tag[dpx_label[TAG_SOURCE_SIZE]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, " ");
    if (2 == list.size())
    {
      source->size[0] = str::str_to_int(list[0]);
      source->size[1] = str::str_to_int(list[1]);
    }
  }
  
  zero(source->file, 100);
  tmp = tag[dpx_label[TAG_SOURCE_FILE]];
  if (tmp.size()) str::ascii(tmp, source->file, 100, false);

  zero(source->time, 24);
  tmp = tag[label[TAG_SOURCE_TIME]];
  if (tmp.size()) str::ascii(tmp, source->time, 24, false);

  zero(source->input_device, 32);
  tmp = tag[dpx_label[TAG_SOURCE_INPUT_DEVICE]];
  if (tmp.size()) str::ascii(tmp, source->input_device, 32, false);

  zero(source->input_serial, 32);
  tmp = tag[dpx_label[TAG_SOURCE_INPUT_SERIAL]];
  if (tmp.size()) str::ascii(tmp, source->input_serial, 32, false);

  tmp = tag[dpx_label[TAG_SOURCE_BORDER]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, " ");
    if (4 == list.size())
    {
      source->border[0] = str::str_to_int(list[0]);
      source->border[1] = str::str_to_int(list[1]);
      source->border[2] = str::str_to_int(list[2]);
      source->border[3] = str::str_to_int(list[3]);
    }
  }

  tmp = tag[dpx_label[TAG_SOURCE_PIXEL_ASPECT]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, " ");
    if (2 == list.size())
    {
      source->pixel_aspect[0] = str::str_to_int(list[0]);
      source->pixel_aspect[1] = str::str_to_int(list[1]);
    }
  }

  tmp = tag[dpx_label[TAG_SOURCE_SCAN_SIZE]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, " ");
    if (2 == list.size())
    {
      source->scan_size[0] = str::str_to_float(list[0]);
      source->scan_size[1] = str::str_to_float(list[1]);
    }
  }
  
  // Film.

  zero(film->id, 2);
  zero(film->type, 2);
  zero(film->offset, 2);
  zero(film->prefix, 6);
  zero(film->count, 4);
  tmp = tag[label[tag::KEYCODE]];
  if (tmp.size())
  {
    int id = 0, type = 0, prefix = 0, count = 0, offset = 0;
    time::str_to_keycode(tmp, &id, &type, &prefix, &count, &offset);
    str::ascii(str::int_to_str(id), film->id, 2, false);
    str::ascii(str::int_to_str(type), film->type, 2, false);
    str::ascii(str::int_to_str(offset), film->offset, 2, false);
    str::ascii(str::int_to_str(prefix), film->prefix, 6, false);
    str::ascii(str::int_to_str(count), film->count, 4, false);
  }
  
  zero(film->format, 32);
  tmp = tag[dpx_label[TAG_FILM_FORMAT]];
  if (tmp.size()) str::ascii(tmp, film->format, 32, false);

  tmp = tag[dpx_label[TAG_FILM_FRAME]];
  if (tmp.size()) film->frame = str::str_to_int(tmp);

  tmp = tag[dpx_label[TAG_FILM_SEQUENCE]];
  if (tmp.size()) film->sequence = str::str_to_int(tmp);

  tmp = tag[dpx_label[TAG_FILM_HOLD]];
  if (tmp.size()) film->hold = str::str_to_int(tmp);

  tmp = tag[dpx_label[TAG_FILM_FRAME_RATE]];
  if (tmp.size()) film->frame_rate = str::str_to_float(tmp);

  tmp = tag[dpx_label[TAG_FILM_SHUTTER]];
  if (tmp.size()) film->shutter = str::str_to_float(tmp);

  zero(film->frame_id, 32);
  tmp = tag[dpx_label[TAG_FILM_FRAME_ID]];
  if (tmp.size()) str::ascii(tmp, film->frame_id, 32, false);

  zero(film->slate, 100);
  tmp = tag[dpx_label[TAG_FILM_SLATE]];
  if (tmp.size()) str::ascii(tmp, film->slate, 100, false);
  
  // TV.

  tmp = tag[label[tag::TIMECODE]];
  if (tmp.size()) tv->timecode = time::str_to_timecode(tmp);

  tmp = tag[dpx_label[TAG_TV_INTERLACE]];
  if (tmp.size()) tv->interlace = str::str_to_int(tmp);

  tmp = tag[dpx_label[TAG_TV_FIELD]];
  if (tmp.size()) tv->field = str::str_to_int(tmp);

  tmp = tag[dpx_label[TAG_TV_VIDEO_SIGNAL]];
  if (tmp.size()) tv->video_signal = str::str_to_int(tmp);
  
  tmp = tag[dpx_label[TAG_TV_SAMPLE_RATE]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, " ");
    if (2 == list.size())
    {
      tv->sample_rate[0] = str::str_to_float(list[0]);
      tv->sample_rate[1] = str::str_to_float(list[1]);
    }
  }
  
  tmp = tag[dpx_label[TAG_TV_FRAME_RATE]];
  if (tmp.size()) tv->frame_rate = str::str_to_float(tmp);

  tmp = tag[dpx_label[TAG_TV_TIME_OFFSET]];
  if (tmp.size()) tv->time_offset = str::str_to_float(tmp);

  tmp = tag[dpx_label[TAG_TV_GAMMA]];
  if (tmp.size()) tv->gamma = str::str_to_float(tmp);

  tmp = tag[dpx_label[TAG_TV_BLACK_LEVEL]];
  if (tmp.size()) tv->black_level = str::str_to_float(tmp);

  tmp = tag[dpx_label[TAG_TV_BLACK_GAIN]];
  if (tmp.size()) tv->black_gain = str::str_to_float(tmp);

  tmp = tag[dpx_label[TAG_TV_BREAK_POINT]];
  if (tmp.size()) tv->breakpoint = str::str_to_float(tmp);

  tmp = tag[dpx_label[TAG_TV_WHITE_LEVEL]];
  if (tmp.size()) tv->white_level = str::str_to_float(tmp);

  tmp = tag[dpx_label[TAG_TV_INTEGRATION_TIMES]];
  if (tmp.size()) tv->integration_times = str::str_to_float(tmp);
}

void _tag(const Header::File & file, const Header::Source & source,
  const Header::Film & film, const Header::Tv & tv, Tag * tag,
  time::Speed * speed)
{
  const List<Str> & label = tag::label();
  const List<Str> & dpx_label = dpx::label_tag();

  // File.
  
  if (valid(file.time, 24))
    (*tag)[label[tag::TIME]] = _str(file.time, 24);
  
  if (valid(file.creator, 100))
    (*tag)[label[tag::CREATOR]] = _str(file.creator, 100);
  
  if (valid(file.project, 200))
    (*tag)[label[tag::PROJECT]] = _str(file.project, 200);
  
  if (valid(file.copyright, 200))
    (*tag)[label[tag::COPYRIGHT]] = _str(file.copyright, 200);
  
  // Source.
  
  if (valid(&source.offset[0]) && valid(&source.offset[1]))
    (*tag)[dpx_label[TAG_SOURCE_OFFSET]] = str::join(List<Str>() <<
      str::int_to_str(source.offset[0]) <<
      str::int_to_str(source.offset[1]), " ");
  
  if (valid(&source.center[0]) && valid(&source.center[1]))
    (*tag)[dpx_label[TAG_SOURCE_CENTER]] = str::join(List<Str>() <<
      str::float_to_str(source.center[0]) <<
      str::float_to_str(source.center[1]), " ");
  
  if (valid(&source.size[0]) && valid(&source.size[1]))
    (*tag)[dpx_label[TAG_SOURCE_SIZE]] = str::join(List<Str>() <<
      str::int_to_str(source.size[0]) <<
      str::int_to_str(source.size[1]), " ");
  
  if (valid(source.file, 100))
    (*tag)[dpx_label[TAG_SOURCE_FILE]] = _str(source.file, 100);
  
  if (valid(source.time, 24))
    (*tag)[label[TAG_SOURCE_TIME]] = _str(source.time, 24);
  
  if (valid(source.input_device, 32))
    (*tag)[dpx_label[TAG_SOURCE_INPUT_DEVICE]] =
      _str(source.input_device, 32);
  
  if (valid(source.input_serial, 32))
    (*tag)[dpx_label[TAG_SOURCE_INPUT_SERIAL]] =
      _str(source.input_serial, 32);
  
  if (valid(&source.border[0]) && valid(&source.border[1]) &&
    valid(&source.border[2]) && valid(&source.border[3])
  )
    (*tag)[dpx_label[TAG_SOURCE_BORDER]] = str::join(List<Str>() <<
      str::int_to_str(source.border[0]) <<
      str::int_to_str(source.border[1]) <<
      str::int_to_str(source.border[2]) <<
      str::int_to_str(source.border[3]), " ");
  
  if (valid(&source.pixel_aspect[0]) && valid(&source.pixel_aspect[1]))
    (*tag)[dpx_label[TAG_SOURCE_PIXEL_ASPECT]] = str::join(List<Str>() <<
      str::int_to_str(source.pixel_aspect[0]) <<
      str::int_to_str(source.pixel_aspect[1]), " ");
  
  if (valid(&source.scan_size[0]) && valid(&source.scan_size[1]))
    (*tag)[dpx_label[TAG_SOURCE_SCAN_SIZE]] = str::join(List<Str>() <<
      str::float_to_str(source.scan_size[0]) <<
      str::float_to_str(source.scan_size[1]), " ");

  // Film.
  
  if (valid(film.id, 2) && valid(film.type, 2) &&
    valid(film.offset, 2) && valid(film.prefix, 6) &&
    valid(film.count, 4))
  {
    (*tag)[label[tag::KEYCODE]] = time::keycode_to_str(
      str::str_to_int(_str(film.id, 2)),
      str::str_to_int(_str(film.type, 2)),
      str::str_to_int(_str(film.prefix, 6)),
      str::str_to_int(_str(film.count, 4)),
      str::str_to_int(_str(film.offset, 2)));
  }
  
  if (valid(film.format, 32))
    (*tag)[dpx_label[TAG_FILM_FORMAT]] = _str(film.format, 32);
  
  if (valid(&film.frame))
    (*tag)[dpx_label[TAG_FILM_FRAME]] = str::int_to_str(film.frame);
  
  if (valid(&film.sequence))
    (*tag)[dpx_label[TAG_FILM_SEQUENCE]] = str::int_to_str(film.sequence);
  
  if (valid(&film.hold))
    (*tag)[dpx_label[TAG_FILM_HOLD]]  = str::int_to_str(film.hold);
  
  if (valid(&film.frame_rate) && film.frame_rate > _speed_min)
  {
    if (speed) *speed = time::float_to_speed(film.frame_rate);
    (*tag)[dpx_label[TAG_FILM_FRAME_RATE]] = str::float_to_str(film.frame_rate);
  }
  
  if (valid(&film.shutter))
    (*tag)[dpx_label[TAG_FILM_SHUTTER]] = str::float_to_str(film.shutter);
  
  if (valid(film.frame_id, 32))
    (*tag)[dpx_label[TAG_FILM_FRAME_ID]] = _str(film.frame_id, 32);
  
  if (valid(film.slate, 100))
    (*tag)[dpx_label[TAG_FILM_SLATE]] = _str(film.slate, 100);

  // TV.
    
  if (valid(&tv.timecode))
    (*tag)[label[tag::TIMECODE]] = time::timecode_to_str(tv.timecode);
  
  if (valid(&tv.interlace))
    (*tag)[dpx_label[TAG_TV_INTERLACE]] = str::int_to_str(tv.interlace);
  
  if (valid(&tv.field))
    (*tag)[dpx_label[TAG_TV_FIELD]] = str::int_to_str(tv.field);
  
  if (valid(&tv.video_signal))
    (*tag)[dpx_label[TAG_TV_VIDEO_SIGNAL]] =
      str::int_to_str(tv.video_signal);
  
  if (valid(&tv.sample_rate[0]) && valid(&tv.sample_rate[1]))
    (*tag)[dpx_label[TAG_TV_SAMPLE_RATE]] = str::join(List<Str>() <<
      str::float_to_str(tv.sample_rate[0]) <<
      str::float_to_str(tv.sample_rate[1]), " ");
  
  if (valid(&tv.frame_rate) && tv.frame_rate > _speed_min)
  {
    if (speed) *speed = time::float_to_speed(tv.frame_rate);
    (*tag)[dpx_label[TAG_TV_FRAME_RATE]] = str::float_to_str(tv.frame_rate);
  }
  
  if (valid(&tv.time_offset))
    (*tag)[dpx_label[TAG_TV_TIME_OFFSET]] = str::float_to_str(tv.time_offset);
  
  if (valid(&tv.gamma))
    (*tag)[dpx_label[TAG_TV_GAMMA]] = str::float_to_str(tv.gamma);
  
  if (valid(&tv.black_level))
    (*tag)[dpx_label[TAG_TV_BLACK_LEVEL]] = str::float_to_str(tv.black_level);
  
  if (valid(&tv.black_gain))
    (*tag)[dpx_label[TAG_TV_BLACK_GAIN]] = str::float_to_str(tv.black_gain);
  
  if (valid(&tv.breakpoint))
    (*tag)[dpx_label[TAG_TV_BREAK_POINT]] = str::float_to_str(tv.breakpoint);
  
  if (valid(&tv.white_level))
    (*tag)[dpx_label[TAG_TV_WHITE_LEVEL]] = str::float_to_str(tv.white_level);
  
  if (valid(&tv.integration_times))
    (*tag)[dpx_label[TAG_TV_INTEGRATION_TIMES]] =
      str::float_to_str(tv.integration_times);
}

}

void header_load(djv::Io * io, Info * info, Tag * tag, time::Speed * speed,
  bool * print) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "header_load");
  
  Header header;
  io->get(&header, sizeof(Header));
  DJV_DBG_PRN("magic = " << Str((char *)&header.file.magic, 4));
  if (0 == mem::compare(&header.file.magic, magic[0], 4))
    info->endian = mem::MSB;
  else if (0 == mem::compare(&header.file.magic, magic[1], 4))
    info->endian = mem::LSB;
  else
    throw Error(name, error_unrecognized);
  if (info->endian != mem::endian())
  {
    DJV_DBG_PRN("endian");
    io->endian(true);
    _endian(&header);
  }
  dbg(header);
  
  _get(header.image, info, print);
  _tag(header.file, header.source, header.film, header.tv, tag, speed);

  if (header.file.image_offset)
    io->position(header.file.image_offset);
}

void header_save(djv::Io * io, const Str & file, const Info & info,
  const Tag & tag, ENDIAN endian, COLOR color) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "header_save");
  DJV_DBG_PRN("info = " << info);
  DJV_DBG_PRN("color = " << color);
  
  Header header;
  mem::set(0xff, &header, sizeof(Header));
  _set(info, color, &header.file, &header.image);
  _tag(file, tag, &header.file, &header.source, &header.film, &header.tv);
  dbg(header);
  
  mem::ENDIAN file_endian = mem::endian();
  if (MSB == endian) file_endian = mem::MSB;
  else if (LSB == endian) file_endian = mem::LSB;
  if (file_endian != mem::endian())
  {
    DJV_DBG_PRN("endian");
    io->endian(true);
    _endian(&header);
  }
  mem::copy(mem::MSB == file_endian ? magic[0] : magic[1],
    &header.file.magic, 4);
  io->set(&header, sizeof(Header));
}

void header_save_update(djv::Io * io) throw (Error)
{
  const size_t size = io->position();
  io->position(12);
  io->set_u32(size);
}

}
}
}
}

