/*------------------------------------------------------------------------------
 lib/djv_image/io_dpx_save.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_dpx_save.cxx"

#include "io_dpx.h"

namespace djv {
namespace image {
namespace io {
namespace dpx
{

Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* dpx color (value) - Output color space: %1. Default: %2.\n"
"* dpx print (black) (white) (gamma) - Print values. Default: %3.\n"
"* dpx type (value) - Pixel type: %4. Default: %5.\n"
"* dpx endian (value) - Endian: %6. Default: %7.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_color(), ", ").
    arg(str::serialize(_options.color, label_color())).
    arg(serialize(_options.print)).
    arg(label_type(), ", ").
    arg(str::serialize(_options.type, label_type())).
    arg(label_endian(), ", ").
    arg(str::serialize(_options.endian, label_endian()));
}

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_COLOR]))
    return str::serialize(data, &_options.color, label_color());
  else if (str::compare(in, option_list()[OPTION_PRINT]))
    return serialize(data, &_options.print);
  else if (str::compare(in, option_list()[OPTION_TYPE]))
    return str::serialize(data, &_options.type, label_type());
  else if (str::compare(in, option_list()[OPTION_ENDIAN]))
    return str::serialize(data, &_options.endian, label_endian());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COLOR]))
    return str::serialize(_options.color, label_color());
  else if (str::compare(in, option_list()[OPTION_PRINT]))
    return serialize(_options.print);
  else if (str::compare(in, option_list()[OPTION_TYPE]))
    return str::serialize(_options.type, label_type());
  else if (str::compare(in, option_list()[OPTION_ENDIAN]))
    return str::serialize(_options.endian, dpx::label_endian());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const Info & info, const Tag &, const Time &)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("info = " << info);
  
  _file = in;
  
  _info = Info();
  _info.size = info.size;
  _info.mirror_y = true;
  switch (_options.endian)
  {
    case ENDIAN_AUTO: break;
    case MSB: _info.endian = mem::MSB; break;
    case LSB: _info.endian = mem::LSB; break;
  }
  switch (_options.type)
  {
    case U10: _info.pixel = pixel::RGB_U10; break;
    default:
    {
      pixel::TYPE type = pixel::type(info.pixel);
      switch (type)
      {
        case pixel::F16: 
        case pixel::F32: type = pixel::U16; break;
        default: break;
      }
      _info.pixel = pixel::pixel(pixel::format(info.pixel), type);
    } break;
  }
  switch (pixel::bit_depth(_info.pixel))
  {
    case 8:
    case 10: _info.align = 4; break;
    default: break;
  }
  DJV_DBG_PRN("info = " << _info);

  _image.set(_info);
}

void Save::_open(const File & in, const Tag & tag) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  _io.open(in, djv::io::WRITE);
  
  header_save(&_io, in, _info, tag, _options.endian, _options.color);
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  
  // Setup.
  
  Profile profile;
  if (dpx::COLOR_PRINT == _options.color || dpx::COLOR_AUTO == _options.color)
  {
    DJV_DBG_PRN("profile");
    profile.type = profile::LUT;
    profile.lut = print_lut(_options.print);
  }

  // Open.
  
  _open(_file.get(frame), in.tag);
  
  // Write.

  const Data * p = &in;
  if (in.info() != _info || in.profile.type != profile::RAW)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::Draw draw;
    draw.profile = profile;
    gl::data::copy(*p, &_image, draw);
    p = &_image;
  }
  
  _io.set(p->data(), p->data_bytes());
  
  header_save_update(&_io);
  _io.close();
}

}
}
}
}

