/*------------------------------------------------------------------------------
 lib/djv_image/io_ppm_load.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_ppm_load.cxx"

#include "io_ppm.h"

namespace djv {
namespace image {
namespace io {
namespace ppm
{

Load::Load()
{
  _io.endian(mem::endian() != mem::MSB);
}

void Load::open(const File & in, Info * info, Tag *, Time * time) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(in.get(in.seq.start), info);
  
  if (file::SEQ == _file.type)
  {
    time->start = _file.seq.start;
    time->end = _file.seq.end;
  }
}

void Load::_open(const File & in, Info * info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::_open");
  DJV_DBG_PRN("in = " << in);
  
  // Open.
  
  _io.open(in, djv::io::READ);
  
  char magic [] = { 0, 0, 0 };
  _io.get(magic, 2);
  if (magic[0] != 'P')
    throw Error(name(), error_unrecognized);
  switch (magic[1])
  {
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6': break;
    default: throw Error(name(), error_unsupported);
  }
  const int ppm_type = magic[1] - '0';
  
  // Header.

  char tmp [str::cstr_size];

  int width, height;
  djv::io::word(&_io, tmp, str::cstr_size);
  width = str::str_to_int(tmp, str::cstr_size);
  djv::io::word(&_io, tmp, str::cstr_size);
  height = str::str_to_int(tmp, str::cstr_size);

  int max_value = 0;
  if (ppm_type != 1 && ppm_type != 4)
  {
    djv::io::word(&_io, tmp, str::cstr_size);
    max_value = str::str_to_int(tmp, str::cstr_size);
  }
  DJV_DBG_PRN("max value = " << max_value);
  
  // Information.
  
  info->size = Vec2i(width, height);
  info->mirror_y = true;

  if (1 == ppm_type || 4 == ppm_type)
    _bit_depth = 1;
  else
    _bit_depth = max_value < 256 ? 8 : 16;
  DJV_DBG_PRN("bit depth = " << _bit_depth);
  
  int channels = 0;
  switch (ppm_type)
  {
    case 1:
    case 2:
    case 4:
    case 5: channels = 1; break;
    case 3:
    case 6: channels = 3; break;
  }
  if (! pixel::pixel(channels, _bit_depth != 1 ? _bit_depth : 8, &info->pixel))
    throw Error(name(), error_unsupported);

  _mode = (1 == ppm_type || 2 == ppm_type || 3 == ppm_type) ? ASCII : BINARY;
  if (BINARY == _mode) info->endian = mem::MSB;
}

const Image * Load::get(int frame, int, int proxy) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Open.
  
  Info info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start), &info);
  
  // Read.
  
  Data * data = proxy ? &_tmp : &_image;

  if (BINARY == _mode && _bit_depth != 1)
  {
    const uint8_t * p = _io.mmap_p();
    _io.seek(data::data_bytes(info));
    data->set(info, p);
  }
  else
  {
    data->set(info);
    const int channels = pixel::channels(info.pixel);
    if (BINARY == _mode && 1 == _bit_depth)
    {
      const size_t scanline_bytes = ppm::scanline_bytes(
        info.size.x, channels, _bit_depth, _mode);
      mem::Buffer<uint8_t> scanline(scanline_bytes);
      DJV_DBG_PRN("scanline bytes = " << scanline_bytes);

      for (int y = 0; y < info.size.y; ++y)
      {
        _io.get(scanline(), scanline_bytes);
        const uint8_t * in_p = scanline();
        uint8_t * out_p = data->data(0, y);
        for (int i = info.size.x - 1; i >= 0; --i)
          out_p[i] = (in_p[i / 8] >> (7 - (i % 8))) & 1 ? 0 : 255;
      }
    }
    else
    {
      for (int y = 0; y < info.size.y; ++y)
        ascii_unpack(&_io, data->data(0, y), info.size.x * channels,
          _bit_depth);
    }
  }
  
  if (proxy)
  {
    info.size = data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _image.set(info);
    data::proxy_scale(_tmp, &_image, proxy);
  }

  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

}
}
}
}
