/*------------------------------------------------------------------------------
 lib/djv_image/io_ppm_save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_ppm_save.cxx"

#include "io_ppm.h"

namespace djv {
namespace image {
namespace io {
namespace ppm
{

Save::Save()
{
  _io.endian(mem::endian() != mem::MSB);
}

Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* ppm type (value) - Pixel type: %1. Default: %2.\n"
"* ppm mode (value) - File mode: %3. Default: %4.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_type(), ", ").arg(str::serialize(_options.type, label_type())).
    arg(label_mode(), ", ").arg(str::serialize(_options.mode, label_mode()));
}

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_TYPE]))
    return str::serialize(data, &_options.type, label_type());
  else if (str::compare(in, option_list()[OPTION_MODE]))
    return str::serialize(data, &_options.mode, label_mode());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_TYPE]))
    return str::serialize(_options.type, label_type());
  else if (str::compare(in, option_list()[OPTION_MODE]))
    return str::serialize(_options.mode, label_mode());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const Info & info, const Tag &, const Time &)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  _info = Info();
  _info.size = info.size;
  _info.mirror_y = true;
  switch (_options.type)
  {
    case TYPE_AUTO:
    {
      pixel::FORMAT format = pixel::format(info.pixel);
      switch (format)
      {
        case pixel::LA: format = pixel::L; break;
        case pixel::RGBA: format = pixel::RGB; break;
        default: break;
      }
      pixel::TYPE type = pixel::type(info.pixel);
      switch (type)
      {
        case pixel::U10:
        case pixel::F16: 
        case pixel::F32: type = pixel::U16; break;
        default: break;
      }
      _info.pixel = pixel::pixel(format, type);
      _bit_depth = pixel::bit_depth(_info.pixel);
    } break;
    case U1: _info.pixel = pixel::L_U8; _bit_depth = 1; break;
  }
  _info.endian = mem::MSB;
  DJV_DBG_PRN("info = " << _info);

  _image.set(_info);
}

void Save::_open(const File & in) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  // Open.
  
  _io.open(in, djv::io::WRITE);
  
  // Header.
  
  int ppm_type = 0;
  if (1 == _bit_depth)
  {
    ppm_type = ASCII == _options.mode ? 1 : 4;
  }
  else
  {
    ppm_type = ASCII == _options.mode ? 2 : 5;
    if (3 == _image.channels()) ++ppm_type;
  }
  char magic [] = "P \n";
  magic[1] = '0' + ppm_type;
  _io.set(magic, 3);
  
  char tmp [str::cstr_size];
  int size = SNPRINTF(tmp, str::cstr_size, "%d %d\n", _image.w(), _image.h());
  _io.set(tmp, size);
  
  if (_bit_depth != 1)
  {
    const int max_value = (8 == _bit_depth) ? pixel::u8_max : pixel::u16_max;
    size = SNPRINTF(tmp, str::cstr_size, "%d\n", max_value);
    _io.set(tmp, size);
  }
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("type = " << _options.type);
  DJV_DBG_PRN("mode = " << _options.mode);
  
  // Open.

  _open(_file.get(frame));
  
  // Write.
  
  const Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    p = &_image;
  }
  
  if (BINARY == _options.mode && _bit_depth != 1)
  {
    _io.set(p->data(), p->data_bytes());
  }
  else
  {
    const int w = p->w(), h = p->h();
    const int channels = pixel::channels(p->info().pixel);
    const size_t scanline_bytes = ppm::scanline_bytes(
      w, channels, _bit_depth, _options.mode);
    mem::Buffer<uint8_t> scanline(scanline_bytes);
    DJV_DBG_PRN("scanline bytes = " << scanline_bytes);
  
    for (int y = 0; y < h; ++y)
      if (BINARY == _options.mode && 1 == _bit_depth)
      {
        const uint8_t * in_p = p->data(0, y);
        uint8_t * out_p = scanline();
        for (int i = 0; i < w; ++i)
        {
          const int tmp = in_p[i];
          const int off = i % 8;
          const int j = i / 8;
          if (0 == off) out_p[j] = 0;
          out_p[j] |= ((! tmp) & 1) << (7 - off);
        }
        _io.set(scanline(), scanline_bytes);
      }
      else
      {
        const size_t size = ascii_pack(
          p->data(0, y), scanline(), w * channels, _bit_depth);
        _io.set(scanline(), size);
      }
  }
  
  _io.close();
}

}
}
}
}

