/*------------------------------------------------------------------------------
 lib/djv_image/io_tga_load.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_tga_load.cxx"

#include "io_tga.h"

namespace djv {
namespace image {
namespace io {
namespace tga
{

Load::Load()
{
  _io.endian(mem::endian() != mem::LSB);
}

void Load::open(const File & in, Info * info, Tag *, Time * time) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start), info);

  if (file::SEQ == _file.type)
  {
    time->start = _file.seq.start;
    time->end = _file.seq.end;
  }
}

void Load::_open(const File & in, Info * info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::_open");
  DJV_DBG_PRN("in = " << in);
  
  // Open.
  
  _io.open(in, djv::io::READ);
  
  // Header.

  uint8_t id_size, cmap_type, image_type;
  uint16_t cmap_start, cmap_size;
  uint8_t cmap_bits;
  uint16_t x, y, width, height;
  uint8_t pixel_bits, descriptor;
  _io.get_u8(&id_size);
  _io.get_u8(&cmap_type);
  _io.get_u8(&image_type);
  _io.get_u16(&cmap_start);
  _io.get_u16(&cmap_size);
  _io.get_u8(&cmap_bits);
  _io.get_u16(&x);
  _io.get_u16(&y);
  _io.get_u16(&width);
  _io.get_u16(&height);
  _io.get_u8(&pixel_bits);
  _io.get_u8(&descriptor);
  _io.seek(id_size);
  DJV_DBG_PRN("id size = " << id_size);
  DJV_DBG_PRN("cmap type = " << cmap_type);
  DJV_DBG_PRN("image type = " << image_type);
  DJV_DBG_PRN("cmap start = " << cmap_start);
  DJV_DBG_PRN("cmap size = " << cmap_size);
  DJV_DBG_PRN("cmap bits = " << cmap_bits);
  DJV_DBG_PRN("image position = " << x << " " << y);
  DJV_DBG_PRN("image size = " << width << " " << height);
  DJV_DBG_PRN("pixel bits = " << pixel_bits);
  
  // Information.
  
  info->size = Vec2i(width, height);
  info->mirror_x = (descriptor >> 4) & 1;
  info->mirror_y = (descriptor >> 5) & 1;

  const int alpha_bits = descriptor & 15;
  DJV_DBG_PRN("alpha bits = " << alpha_bits);
  switch (alpha_bits)
  {
    case 0:
    case 8: break;
    default: throw Error(name(), error_unsupported);
  }

  int pixel = -1;
  switch (image_type)
  {
    case 3:
    case 11:
      switch (pixel_bits)
      {
        case 8: pixel = pixel::L_U8; break;
        case 16: pixel = pixel::LA_U8; break;
      }
      break;
    case 2:
    case 10:
      switch (pixel_bits)
      {
        case 24: pixel = pixel::RGB_U8; info->color = image::data::BGR; break;
        case 32: pixel = pixel::RGBA_U8; info->color = image::data::BGR; break;
      }
      break;
  }
  if (-1 == pixel)
    throw Error(name(), error_unsupported);
  info->pixel = PIXEL(pixel);
  DJV_DBG_PRN("pixel = " << info->pixel);
  
  const int bits = pixel_bits + alpha_bits;
  if (bits < (pixel::channels(info->pixel) * 8) || (bits % 8) != 0)
    throw Error(name(), error_unsupported);
  
  _compress = 10 == image_type || 11 == image_type;
  DJV_DBG_PRN("compress = " << _compress);
  
  info->endian = mem::LSB;
}

const Image * Load::get(int frame, int, int proxy) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Open.
  
  Info info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start), &info);

  // Read.
  
  Data * data = proxy ? &_tmp : &_image;
  
  if (! _compress)
  {
    const uint8_t * p = _io.mmap_p();
    _io.seek(data::data_bytes(info));
    data->set(info, p);
  }
  else
  {
    data->set(info);
    const int channels = pixel::channels(info.pixel);
    for (int y = 0; y < info.size.y; ++y)
    {
      DJV_DBG_PRN("y = " << y);
      rle_unpack(&_io, data->data(0, y), info.size.x, channels);
    }
  }
  
  if (proxy)
  {
    info.size = data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _image.set(info);
    data::proxy_scale(_tmp, &_image, proxy);
  }
  
  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

}
}
}
}
