/*------------------------------------------------------------------------------
 lib/djv_image/pixel.h
 
 Pixels
 ------
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_PIXEL_H
#define DJV_IMAGE_PIXEL_H

#include <djv_image/base.h>
#include <half.h>

namespace djv {
namespace image {
namespace pixel
{

//------------------------------------------------------------------------------
// Pixel Components

enum FORMAT { L, LA, RGB, RGBA };
enum TYPE { U8, U10, U16, F16, F32 };

typedef uint8_t  U8_T;
typedef uint16_t U10_T;
typedef uint16_t U16_T;
typedef half     F16_T;
typedef float    F32_T;

#if defined(DJV_MSB) // XXX
struct U10_S { uint r:10, g:10, b:10, pad:2; };
#else
struct U10_S { uint pad:2, b:10, g:10, r:10; };
#endif

static const int channels_max = 4;
static const int bytes_max = 4;

static const int u8_range  = 256;
static const int u10_range = 1024;
static const int u16_range = 65536;

static const int u8_max  = u8_range - 1;
static const int u10_max = u10_range - 1;
static const int u16_max = u16_range - 1;

//------------------------------------------------------------------------------
// Pixel Type

enum PIXEL
{
  L_U8, L_U16, L_F16, L_F32,
  LA_U8, LA_U16, LA_F16, LA_F32,
  RGB_U8, RGB_U10, RGB_U16, RGB_F16, RGB_F32,
  RGBA_U8, RGBA_U16, RGBA_F16, RGBA_F32
};

static const int pixel_size = 17;

//------------------------------------------------------------------------------
// Functions

inline FORMAT format(PIXEL);
inline TYPE type(PIXEL);

inline int channels(FORMAT);
inline int channels(PIXEL);
inline int channel_bytes(PIXEL);
inline int bytes(PIXEL);
inline int bit_depth(PIXEL);
inline int range(PIXEL);
inline int max(PIXEL);

inline PIXEL pixel(FORMAT, TYPE);
inline bool pixel(FORMAT, TYPE, PIXEL *);
inline bool pixel(int channels, int bit_depth, PIXEL *);
inline int pixel(int channels, int bit_depth);

//------------------------------------------------------------------------------
// Conversion

inline U10_T u8_to_u10(U8_T);
inline U16_T u8_to_u16(U8_T);
inline F16_T u8_to_f16(U8_T);
inline F32_T u8_to_f32(U8_T);

inline U8_T  u10_to_u8 (U10_T);
inline U16_T u10_to_u16(U10_T);
inline F16_T u10_to_f16(U10_T);
inline F32_T u10_to_f32(U10_T);

inline U8_T  u16_to_u8 (U16_T);
inline U10_T u16_to_u10(U16_T);
inline F16_T u16_to_f16(U16_T);
inline F32_T u16_to_f32(U16_T);

inline U8_T  f16_to_u8 (F16_T);
inline U10_T f16_to_u10(F16_T);
inline U16_T f16_to_u16(F16_T);
inline F32_T f16_to_f32(F16_T);

inline U8_T  f32_to_u8 (F32_T);
inline U10_T f32_to_u10(F32_T);
inline U16_T f32_to_u16(F32_T);
inline F16_T f32_to_f16(F32_T);

#define PIXEL_U8_ZERO  0
#define PIXEL_U10_ZERO 0
#define PIXEL_U16_ZERO 0
#define PIXEL_F16_ZERO 0.0
#define PIXEL_F32_ZERO 0.0

#define PIXEL_U8_ONE   pixel::u8_max
#define PIXEL_U10_ONE  pixel::u10_max
#define PIXEL_U16_ONE  pixel::u16_max
#define PIXEL_F16_ONE  1.0
#define PIXEL_F32_ONE  1.0

#define PIXEL_U8_TO_U8(IN)   (IN)
#define PIXEL_U8_TO_U10(IN)  pixel::u8_to_u10(IN)
#define PIXEL_U8_TO_U16(IN)  pixel::u8_to_u16(IN)
#define PIXEL_U8_TO_F16(IN)  pixel::u8_to_f16(IN)
#define PIXEL_U8_TO_F32(IN)  pixel::u8_to_f32(IN)

#define PIXEL_U10_TO_U8(IN)  pixel::u10_to_u8(IN)
#define PIXEL_U10_TO_U10(IN) (IN)
#define PIXEL_U10_TO_U16(IN) pixel::u10_to_u16(IN)
#define PIXEL_U10_TO_F16(IN) pixel::u10_to_f16(IN)
#define PIXEL_U10_TO_F32(IN) pixel::u10_to_f32(IN)

#define PIXEL_U16_TO_U8(IN)  pixel::u16_to_u8(IN)
#define PIXEL_U16_TO_U10(IN) pixel::u16_to_u10(IN)
#define PIXEL_U16_TO_U16(IN) (IN)
#define PIXEL_U16_TO_F16(IN) pixel::u16_to_f16(IN)
#define PIXEL_U16_TO_F32(IN) pixel::u16_to_f32(IN)

#define PIXEL_F16_TO_U8(IN)  pixel::f16_to_u8(IN)
#define PIXEL_F16_TO_U10(IN) pixel::f16_to_u10(IN)
#define PIXEL_F16_TO_U16(IN) pixel::f16_to_u16(IN)
#define PIXEL_F16_TO_F16(IN) (IN)
#define PIXEL_F16_TO_F32(IN) pixel::f16_to_f32(IN)

#define PIXEL_F32_TO_U8(IN)  pixel::f32_to_u8(IN)
#define PIXEL_F32_TO_U10(IN) pixel::f32_to_u10(IN)
#define PIXEL_F32_TO_U16(IN) pixel::f32_to_u16(IN)
#define PIXEL_F32_TO_F16(IN) pixel::f32_to_f16(IN)
#define PIXEL_F32_TO_F32(IN) (IN)

void convert(const void * in, PIXEL, void * out, PIXEL, bool bgr = false);

bool serialize(List<Str> *, PIXEL *);
List<Str> serialize(const PIXEL &);

//------------------------------------------------------------------------------
// Labels

const List<Str> & label_format();
const List<Str> & label_type();
const List<Str> & label_pixel();

//------------------------------------------------------------------------------
// Debugging

Dbg & operator << (Dbg &, FORMAT);
Dbg & operator << (Dbg &, TYPE);
Dbg & operator << (Dbg &, PIXEL);

}
}
}

//------------------------------------------------------------------------------
// Declarations

namespace djv {
namespace image
{
using pixel::PIXEL;
}
}

#include "djv_image/pixel_.h"

#endif

