/*------------------------------------------------------------------------------
 lib/djv_ui/choice.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_ui/choice.cxx"

#include "choice.h"
#include <djv_ui/style.h>
#include <djv_ui/icon.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/text.h>
#include <FL/Fl.H>
#include <FL/Fl_Menu_Button.H>
#include <FL/fl_draw.H>

namespace djv {
namespace ui {
namespace choice
{

class Widget::_Widget : public Fl_Menu_Button, public widget::Base
{
public:

  _Widget(const List<Str> &);
  
  void set(int);
  int get() const { return _value; }
  Signal<int> sig;
  
  void list(const List<Str> &);
  const List<Str> & list() const { return _list; }

  int handle(int);
  void draw();
  
  void dirty();
  void geom(const Box2i &);
  const Box2i & geom() const { return widget::Base::geom(); }
  void resize(int x, int y, int w, int h) { geom(Box2i(x, y, w, h)); }
  int visible() const { return Fl_Menu_Button::visible(); }
  
  Str dbg_name() const { return "djv::ui::choice::Widget::_Widget"; }

private:

  DJV_FL_WIDGET_CALLBACK(_Widget, menu_callback) { set(value()); }

  void widget_update();

  int _value;
  List<Str> _list;
  Fl_Image * _icon;
};

Widget::_Widget::_Widget(const List<Str> & list) :
  Fl_Menu_Button(0, 0, 0, 0),
  _value(0), _list(list),
  _icon(ui::icon::get("tool_dec"))
{
  box(FL_THIN_UP_BOX);
  
  widget_update();
  
  callback(menu_callback, this);
}

void Widget::_Widget::set(int in)
{
  if (in == _value) return;
  _value = in;
  redraw();
  sig.emit(_value);
}

void Widget::_Widget::list(const List<Str> & in)
{
  if (in == _list) return;
  _list = in;
  _value = 0;
  widget_update();
  sig.emit(_value);
}

int Widget::_Widget::handle(int in)
{
  switch (widget::Base::handle(in))
  {
    case -1: return 0;
    case 1: return 1;
  }
  return Fl_Menu_Button::handle(in);
}

void Widget::_Widget::draw()
{
  Fl_Menu_Button::draw_box(box(), color());
  fl_draw_box(box(), x(), y(), w(), h(), color());

  const int margin =  Fl::box_dx(box()) + 3;
  const Box2i geom(x() + margin, y() + margin, w() - margin * 2,
    h() - margin * 2);
  
  if (_value >= 0 && _value < int(_list.size()))
  {
    fl_font(textfont(), textsize());
    fl_color(FL_FOREGROUND_COLOR);
    fl_draw(_list[_value].c_str(), geom.x, geom.y + fl_height() - fl_descent());
  }
  
  if (_icon)
    _icon->draw(geom.x + geom.w - style::global()->size_tool_icon(),
      geom.y + geom.h / 2 - _icon->h() / 2);

  if (Fl::focus() == this) draw_focus();
}
  
void Widget::_Widget::dirty()
{
  DJV_DBG_OBJ2(_DBG, "Widget::_Widget::dirty");
  
  widget::Base::dirty();

  textfont(text_font());
  textsize(style::global()->font_size());
  textcolor(text_color());
  
  fl_font(text_font(), style::global()->font_size());
  Vec2i size;
  for (uint i = 0; i < _list.size(); ++i)
    size = max(str_size(_list[i]), size);
  
  size_hint(Vec2i(size.x + style::global()->size_tool_icon() +
    style::global()->margin_widget(), size.y) +
    style::global()->margin_widget() * 2 +
    Fl::box_dx(box()) * 2);
}
  
void Widget::_Widget::geom(const Box2i & in)
{
  Fl_Menu_Button::resize(in.x, in.y, in.w, in.h);
  widget::Base::geom(in);
}

void Widget::_Widget::widget_update()
{
  callbacks(false);
  
  clear();
  const uint size = _list.size();
  mem::Buffer<Fl_Menu_Item> tmp(size + 1);
  mem::set(0, tmp(), (size + 1) * sizeof(Fl_Menu_Item));
  Fl_Menu_Item * p = tmp();
  for (uint i = 0; i < size; ++i)
  {
    p[i].text = _list[i].c_str();
    p[i].callback_ = menu_callback;
    p[i].user_data_ = this;
  }
  copy(tmp());
  update();

  callbacks(true);
}

Widget::Widget(const List<Str> & list) : _label(0)
{
  DJV_DBG_OBJ2(_DBG, "Widget::Widget");
  
  _widget = new _Widget(list);

  size(layout::FIXED, layout::FIXED);
  align(layout::LEFT, layout::CENTER);
  
  _layout = new layout::Horizontal(this);
  _layout->margin(0);
  _layout->add(_widget);

  _widget->sig.set(this, widget_callback);
}

void Widget::set(int in) { _widget->set(in); }
int Widget::get() const { return _widget->get(); }

void Widget::list(const List<Str> & in) { _widget->list(in); }
const List<Str> & Widget::list() const { return _widget->list(); }

void Widget::label(widget::Widget * in)
{
  DJV_DBG_OBJ2(_DBG, "Widget::label");
  
  if (_label) _layout->del(_label);
  _label = in;
  if (_label) _layout->insert(_label, 0);
  //update();
}

void Widget::widget_callback(int in) { sig.emit(in); }

}
}
}
