/*------------------------------------------------------------------------------
 lib/djv_ui/dialog.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "dialog.h"
#include <djv_ui/style.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/group.h>
#include <djv_ui/label.h>
#include <djv_ui/button.h>
#include <djv_ui/text.h>
#include <djv_ui/choice.h>

namespace djv {
namespace ui {
namespace dialog
{

Base::Base(const Str & title) : window::Widget(title) {}
Base::~Base() {}

Message_Base::Message_Base(const Str & title) : Base(title) {}
Message_Base::~Message_Base() {}

void Message_Base::dirty()
{
  Base::dirty();
  
  _size_min = max(Base::size_min(), Vec2i(300, 100));
}

Message::Message(const Str & label) : Message_Base("Message")
{
  // Widgets.

  _widget = new label::Widget(label);
  
  button::Push * close_button = new button::Push("Clos&e");
  
  // Layout.

  layout::Vertical * layout = new layout::Vertical(this);
  
  layout::Horizontal * h_layout = new layout::Horizontal(layout);
  layout->stretch(h_layout);
  h_layout->margin(style::global()->margin_large());
  layout->add(_widget);
  layout->stretch(_widget);
  
  h_layout = new layout::Horizontal(layout);
  h_layout->margin(0);
  h_layout->add_stretch();
  h_layout->add(close_button);
  
  // Initialize.

  size(size_hint());
  
  close_button->take_focus();
  
  // Callbacks.

  close_button->sig.set(this, close_callback);
}

void Message::label(const Str & in)
{
  _widget->set(in);
  size(size_hint());
}

Str Message::label() const { return _widget->get(); }

void Message::close_callback(bool)
{
  hide();
}

Question::Question(const Str & label) : Message_Base("Question")
{
  // Widgets.

  _widget = new label::Widget(label);

  button::Push * ok_button = new button::Push("&Yes");
  button::Push * close_button = new button::Push("&No");

  // Layout.
  
  layout::Vertical * layout = new layout::Vertical(this);
  
  layout::Horizontal * h_layout = new layout::Horizontal(layout);
  layout->stretch(h_layout);
  h_layout->margin(style::global()->margin_large());
  h_layout->add_stretch(false);
  h_layout->add(_widget);
  h_layout->add_stretch(false);
  
  h_layout = new layout::Horizontal(layout);
  h_layout->margin(0);
  h_layout->add_stretch();
  h_layout->add(ok_button);
  h_layout->add(close_button);
  
  // Initialize.

  size(size_hint(), false);
  
  close_button->take_focus();

  // Callbacks.
  
  ok_button->sig.set(this, ok_callback);
  close_button->sig.set(this, close_callback);
}

void Question::label(const Str & in)
{
  _widget->set(in);
  size(size_hint(), false);
}

Str Question::label() const { return _widget->get(); }

void Question::ok_callback(bool)
{
  sig.emit(true);
  hide();
}

void Question::close_callback(bool)
{
  hide();
}

Input::Input(const Str & label, const Str & value) : Message_Base("Input")
{
  // Widgets.

  _label = new label::Widget(label);
  _widget = new text::Edit(value);

  button::Push * ok_button = new button::Push("&Ok");
  button::Push * close_button = new button::Push("&Cancel");

  // Layout.

  layout::Vertical * layout = new layout::Vertical(this);
  
  layout::Horizontal * h_layout = new layout::Horizontal(layout);
  layout->stretch(h_layout);
  h_layout->margin(style::global()->margin_large());
  h_layout->add(_label);
  h_layout->add(_widget);
  h_layout->stretch(_widget);
  
  h_layout = new layout::Horizontal(layout);
  h_layout->margin(0);
  h_layout->add_stretch();
  h_layout->add(ok_button);
  h_layout->add(close_button);
  
  // Initialize.

  size(size_hint());
  
  _widget->take_focus();

  // Callbacks.
  
  _widget->action_sig.set(this, widget_callback);
  ok_button->sig.set(this, ok_callback);
  close_button->sig.set(this, close_callback);
}

void Input::set(const Str & in)
{
  _widget->set(in);
  size(size_hint());
}

Str Input::get() const { return _widget->get(); }

void Input::label(const Str & in)
{
  _label->set(in);
  size(size_hint());
}

Str Input::label() const { return _label->get(); }

void Input::widget_callback(const Str &)
{
  ok_callback(true);
}

void Input::ok_callback(bool)
{
  sig.emit(_widget->get());
  hide();
}

void Input::close_callback(bool)
{
  hide();
}

Choice::Choice(const Str & label, const List<Str> & list) :
  Message_Base("Choice")
{
  // Widgets.

  _widget = new choice::Widget(list);
  _label = new label::Widget(label);
  _widget->label(_label);

  button::Push * ok_button = new button::Push("&Ok");
  button::Push * close_button = new button::Push("&Cancel");

  // Layout.

  layout::Vertical * layout = new layout::Vertical(this);
  
  layout::Horizontal * h_layout = new layout::Horizontal(layout);
  layout->stretch(h_layout);
  h_layout->margin(style::global()->margin_large());
  h_layout->add_stretch(false);
  h_layout->add(_widget);
  h_layout->add_stretch(false);
  
  h_layout = new layout::Horizontal(layout);
  h_layout->margin(0);
  h_layout->add_stretch();
  h_layout->add(ok_button);
  h_layout->add(close_button);
  
  // Initialize.

  size(size_hint(), false);
  
  _widget->take_focus();

  // Callbacks.
  
  ok_button->sig.set(this, ok_callback);
  close_button->sig.set(this, close_callback);
}

void Choice::list(const List<Str> & in)
{
  _widget->list(in);
  size(size_hint(), false);
}

List<Str> Choice::list() const { return _widget->list(); }

void Choice::label(const Str & in)
{
  _label->set(in);
  size(size_hint(), false);
}

Str Choice::label() const { return _label->get(); }

void Choice::ok_callback(bool)
{
  sig.emit(_widget->get());
  hide();
}

void Choice::close_callback(bool)
{
  hide();
}

}
}
}
