/*------------------------------------------------------------------------------
 lib/djv_ui/file_browser_dialog.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_ui/file_browser_dialog.cxx"

#include "_file_browser.h"
#include <djv_base/dir.h>
#include <djv_ui/style.h>
#include <djv_ui/prefs.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/group.h>
#include <djv_ui/button.h>
#include <djv_ui/text.h>
#include <djv_ui/columns.h>
#include <djv_ui/menu.h>

namespace djv {
namespace ui {
namespace file_browser
{

Dialog::Dialog() :
  dialog::Base("File Browser"),
  _images(true), _seq(true), _hidden(false),
  _sort(SORT_TIME), _sort_reverse(true), _sort_dirs(true),
  _pin(false),
  _style_size(style::SIZE(0))
{
  DJV_DBG_OBJ2(_DBG, "Dialog::Dialog");
  
  // Widgets.

  _menu_widget = new menu::Widget;

  _widget = new text::Edit;
  _widget->tooltip("Selection.");
  
  _dir_widget = new text::Edit;
  _dir_widget->tooltip("Current directory.");

  button::Tool * up_button = new button::Tool("dir_up");
  up_button->tooltip("Up a directory.\nKeyboard shortcut: Back Slash.");
  button::Tool * prev_button = new button::Tool("dir_prev");
  prev_button->tooltip("Previous directory.\nKeyboard shortcut: Backspace.");
  button::Tool * reload_button = new button::Tool("dir_reload");
  reload_button->tooltip("Reload directory.\nKeyboard shortcut: R.");

  _dir_browser = new Browser;
  
  _columns_widget = new columns::Widget;
  _columns_widget->name_list(label_sort());
  
  _pin_button = new button::Check("Pin");
  button::Push * ok_button = new button::Push("&Ok");
  button::Push * close_button = new button::Push("Clos&e");
  
  group::Frame * frame = new group::Frame;
  frame->edges(false, false, true, false);
  group::Frame * browser_frame = new group::Frame;
  browser_frame->frame(FL_THIN_DOWN_FRAME);
  
  // Layout.

  layout::Vertical * window_layout = new layout::Vertical(this);
  window_layout->spacing(0);
  window_layout->margin(0);
  window_layout->add(_menu_widget);
  window_layout->add(frame);
  window_layout->stretch(frame);
  layout::Vertical * layout = new layout::Vertical(frame);
  
  layout::Horizontal * h_layout = new layout::Horizontal(layout);
  h_layout->margin(0);
  h_layout->add(_dir_widget);
  h_layout->stretch(_dir_widget);
  layout::Horizontal * tool_layout = new layout::Horizontal(h_layout);
  tool_layout->spacing(0);
  tool_layout->margin(0);
  tool_layout->add(up_button);
  tool_layout->add(prev_button);
  tool_layout->add(reload_button);
  
  layout->add(browser_frame);
  layout->stretch(browser_frame);
  layout::Vertical * v_layout = new layout::Vertical(browser_frame);
  v_layout->spacing(0);
  v_layout->margin(0);
  v_layout->add(_columns_widget);
  v_layout->add(_dir_browser);
  v_layout->stretch(_dir_browser);
  
  layout->add(_widget);
  
  h_layout = new ui::layout::Horizontal(layout);
  h_layout->margin(0);
  h_layout->add(_pin_button);
  h_layout->add_spacer(-1, true);
  h_layout->add(ok_button);
  h_layout->add(close_button);
  
  // Initialize.
  
  prefs::Prefs prefs(prefs::global(), "file_browser");
  List<Str> tmp;
  if (prefs.get("recent", &tmp)) str::serialize(&tmp, &_recent);
  if (prefs.get("images", &tmp)) str::serialize(&tmp, &_images);
  if (prefs.get("sequence", &tmp)) str::serialize(&tmp, &_seq);
  if (prefs.get("hidden", &tmp)) str::serialize(&tmp, &_hidden);
  if (prefs.get("sort", &tmp)) str::serialize(&tmp, &_sort, label_sort());
  if (prefs.get("sort_reverse", &tmp)) str::serialize(&tmp, &_sort_reverse);
  if (prefs.get("sort_dirs", &tmp)) str::serialize(&tmp, &_sort_dirs);
  if (prefs.get("bookmarks", &tmp)) str::serialize(&tmp, &_bookmarks);
  if (prefs.get("pin", &tmp)) str::serialize(&tmp, &_pin);
  
  _dir = djv::file::path_fix(".");
  _dir_prev = _dir;
  
  //dir_update();
  widget_update();
  menu_update();
  _pin_button->set(_pin);
  
  size(Vec2i(550, 450));

  _columns_widget->size_list(columns_default_size_list());
  _dir_browser->column_list(_columns_widget->size_list());
  
  //_browser->take_focus();
  close_button->take_focus();
  
  // Callbacks.

  _widget->sig.set(this, widget_callback);
  _widget->action_sig.set(this, widget_action_callback);
  _dir_widget->sig.set(this, dir_callback);
  _columns_widget->sig.set(this, columns_callback);
  _columns_widget->action_sig.set(this, columns_action_callback);
  _dir_browser->sig.set(this, browser_callback);
  _dir_browser->action_sig.set(this, browser_action_callback);

  up_button->sig.set(this, up_callback);
  prev_button->sig.set(this, prev_callback);
  reload_button->sig.set(this, reload_callback);

  _pin_button->sig.set(this, pin_callback);
  ok_button->sig.set(this, ok_callback);
  close_button->sig.set(this, close_callback);
}

void Dialog::set(const Str & in)
{
  if (in == _value) return;
  
  DJV_DBG_OBJ2(_DBG, "Dialog::set");
  DJV_DBG_PRN("in = " << in);

  _value = in;

  widget_update();
}

File Dialog::get() const
{
  DJV_DBG_OBJ2(_DBG, "Dialog::get");
  DJV_DBG_PRN("value = " << _value);
  DJV_DBG_PRN("dir = " << _dir);
  
  File out = _value;
  
  Str path = out.path();
  if (path.size())
  {
    if (! djv::file::is_path_absolute(path))
      path = _dir + path;
    out.path(djv::file::path_fix(path));
  }
  else
  {
    out.path(_dir);
  }
  
  out.stat();
  if (_seq && out.seq_valid()) out.type = djv::file::SEQ;
  DJV_DBG_PRN("out = " << out);
  return out;
}

void Dialog::dir(const Str & in)
{
  const Str tmp = djv::file::path_fix(in);
  //if (tmp == _dir) return;

  DJV_DBG_OBJ2(_DBG, "Dialog::dir");
  DJV_DBG_PRN("in = " << tmp);
  
  _value = Str();
  _dir_prev = _dir;
  _dir = tmp;
  
  dir_update();
  widget_update();
}

void Dialog::images(bool in)
{
  if (in == _images) return;
  _images = in;
  dir_update();
  menu_update();
}

void Dialog::seq(bool in)
{
  if (in == _seq) return;
  _seq = in;
  dir_update();
  menu_update();
}

void Dialog::hidden(bool in)
{
  if (in == _hidden) return;
  _hidden = in;
  dir_update();
  menu_update();
}

void Dialog::sort(SORT in)
{
  if (in == _sort) return;
  _sort = in;
  dir_update();
  menu_update();
}

void Dialog::sort_reverse(bool in)
{
  if (in == _sort_reverse) return;
  _sort_reverse = in;
  dir_update();
  menu_update();
}

void Dialog::sort_dirs(bool in)
{
  if (in == _sort_dirs) return;
  _sort_dirs = in;
  dir_update();
  menu_update();
}

void Dialog::bookmarks(const List<Str> & in)
{
  if (in == _bookmarks) return;
  _bookmarks = in;
  menu_update();
}

void Dialog::pin(bool in)
{
  if (in == _pin) return;
  _pin = in;
  _pin_button->set(_pin);
}

void Dialog::draw()
{
  dialog::Base::draw();
}

void Dialog::widget_del()
{
  prefs::Prefs prefs(prefs::global(), "file_browser");
  prefs.set("recent", str::serialize(_recent));
  prefs.set("images", str::serialize(_images));
  prefs.set("sequence", str::serialize(_seq));
  prefs.set("hidden", str::serialize(_hidden));
  prefs.set("sort", str::serialize(_sort, label_sort()));
  prefs.set("sort_reverse", str::serialize(_sort_reverse));
  prefs.set("sort_dirs", str::serialize(_sort_dirs));
  prefs.set("bookmarks", str::serialize(_bookmarks));
  prefs.set("pin", str::serialize(_pin));
  
  dialog::Base::widget_del();
}

void Dialog::dirty()
{
  dialog::Base::dirty();

  if (style::global()->size() != _style_size)
  {
    _style_size = style::global()->size();
    dir_update();
  }
}

void Dialog::dir_update()
{
  DJV_DBG_OBJ2(_DBG, "Dialog::dir_update");
  DJV_DBG_PRN("dir = " << _dir);
  
  callbacks(false);
  
  _dir_widget->set(_dir);
  
  ui::Cursor cursor(FL_CURSOR_WAIT);
  
  const Dir dir(_dir);
  List<File> list = dir.list();
  if (_seq) djv::file::seq_compress(&list);
  if (! _hidden) djv::file::filter(&list, djv::file::FILTER_HIDDEN);
  djv::file::SORT sort = djv::file::SORT(0);
  switch (_sort)
  {
    case SORT_NAME: sort = djv::file::SORT_NAME; break;
    case SORT_SIZE: sort = djv::file::SORT_SIZE; break;
#if ! defined(WIN32)
    case SORT_USER: sort = djv::file::SORT_USER; break;
#endif
    case SORT_TIME: sort = djv::file::SORT_TIME; break;
  }
  djv::file::sort(&list, sort, _sort_reverse);
  if (_sort_dirs) list = djv::file::sort_dirs(list);
  if (dir.valid()) list.push_front(File(_dir + "/" + ".."));
  
  _columns_widget->sort(_sort);
  _columns_widget->sort_reverse(_sort_reverse);
  _dir_browser->set(list, _images);

  callbacks(true);
}

void Dialog::widget_update()
{
  DJV_DBG_OBJ2(_DBG, "Dialog::widget_update");
  DJV_DBG_PRN("value = " << _value);
  
  callbacks(false);
  
  _widget->set(_value);
    
  callbacks(true);
}

}
}
}
