/*------------------------------------------------------------------------------
 plugin/djv_quicktime_apple/load.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_quicktime_apple/load.cxx"

#include "plugin.h"
#include <CFString.h>
#include <ImageCompression.h>

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_load() { return new Load; }
}

Load::Load() throw (Error) : _movie(0), _gworld(0)
{
  DJV_DBG_OBJ2(_DBG, "Load::Load");
  
  quicktime_init();
}

Load::~Load()
{
  DJV_DBG_OBJ2(_DBG, "Load::~Load");

  _close();
  quicktime_del();
}

namespace
{
const OSType media_list [] = { VisualMediaCharacteristic };
const int media_list_size = 1;

long _frame_count(::Movie in)
{
  DJV_DBG_OBJ2(_DBG, "_frame_count");
  long out = 0;
  for (::TimeValue time = 0; time >= 0; ++out)
  {
    ::TimeValue tmp = 0;
    ::GetMovieNextInterestingTime(in, ::nextTimeStep, media_list_size,
      media_list, time, fixed1,  &tmp, 0);
    //DJV_DBG_PRN("time = " << int(tmp - time));
    time = tmp;
  }
  DJV_DBG_PRN("out = " << int(out));
  return out;
}

::TimeValue _frame_duration(::Movie in)
{
  ::TimeValue out = 0;
  ::GetMovieNextInterestingTime(in, ::nextTimeStep, media_list_size, media_list,
    out, fixed1, 0, &out);
  return out;
}
}

void Load::open(const File & in, image::Info * info, image::Tag * tag,
  image::Time * time) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _close();

  // Open.

  class _CFStringRef
  {
  public:
    _CFStringRef(const Str & in) :
      _p(::CFStringCreateWithCString(0, in.c_str(), 0)) {}
    ~_CFStringRef() { ::CFRelease(_p); }
  private:
    ::CFStringRef _p;
  } file(in);

  ::QTNewMoviePropertyElement prop_list [10];
  ::ItemCount prop_size = 0;

  prop_list[prop_size].propClass = kQTPropertyClass_DataLocation;
  prop_list[prop_size].propID = kQTDataLocationPropertyID_CFStringNativePath;
  prop_list[prop_size].propValueSize = sizeof(file);
  prop_list[prop_size].propValueAddress = (void*)&file;
  prop_list[prop_size].propStatus = 0;
  ++prop_size;

  bool unresolved = true;
  prop_list[prop_size].propClass = kQTPropertyClass_MovieInstantiation;
  prop_list[prop_size].propID =
    kQTMovieInstantiationPropertyID_DontAskUnresolvedDataRefs;
  prop_list[prop_size].propValueSize = sizeof(unresolved);
  prop_list[prop_size].propValueAddress = &unresolved;
  prop_list[prop_size].propStatus = 0;
  ++prop_size;

  bool active = true;
  prop_list[prop_size].propClass = kQTPropertyClass_NewMovieProperty;
  prop_list[prop_size].propID = kQTNewMoviePropertyID_Active;
  prop_list[prop_size].propValueSize = sizeof(active);
  prop_list[prop_size].propValueAddress = &active;
  prop_list[prop_size].propStatus = 0;
  ++prop_size;

  bool no_interact = true;
  prop_list[prop_size].propClass = kQTPropertyClass_NewMovieProperty;
  prop_list[prop_size].propID = kQTNewMoviePropertyID_DontInteractWithUser;
  prop_list[prop_size].propValueSize = sizeof(no_interact);
  prop_list[prop_size].propValueAddress = &no_interact;
  prop_list[prop_size].propStatus = 0;
  ++prop_size;

  QTVisualContextRef visual = 0;
  prop_list[prop_size].propClass = kQTPropertyClass_Context;
  prop_list[prop_size].propID = kQTContextPropertyID_VisualContext;
  prop_list[prop_size].propValueSize = sizeof(QTVisualContextRef);
  prop_list[prop_size].propValueAddress = &visual;
  prop_list[prop_size].propStatus = 0;
  ++prop_size;

  QTAudioContextRef  audio = 0;
  prop_list[prop_size].propClass = kQTPropertyClass_Context;
  prop_list[prop_size].propID = kQTContextPropertyID_AudioContext;
  prop_list[prop_size].propValueSize = sizeof(QTAudioContextRef);
  prop_list[prop_size].propValueAddress = &audio;
  prop_list[prop_size].propStatus = 0;
  ++prop_size;

  OSErr err = ::NewMovieFromProperties(prop_size, prop_list, 0, 0, &_movie);
  if (err != noErr)
    throw Error(::name, str::Format(error_open).arg(err));
  
  // Information.
  
  ::Rect rect;
  ::GetMovieBox(_movie, &rect);
  _info.size = Vec2i(rect.right - rect.left, rect.bottom - rect.top);
  _info.mirror_y = true;
  _info.pixel = image::pixel::RGBA_U8;
  
  const ::TimeScale time_scale = ::GetMovieTimeScale(_movie);
  const ::TimeValue time_duration = ::GetMovieDuration(_movie);
  const long frame_count = _frame_count(_movie);
  const ::TimeValue frame_duration = _frame_duration(_movie);
  //const TimeValue frame_duration = time_duration / frame_count;
  DJV_DBG_PRN("time scale = " << int(time_scale));
  DJV_DBG_PRN("time duration = " << int(time_duration));
  DJV_DBG_PRN("frame count = " << int(frame_count));
  DJV_DBG_PRN("frame duration = " << int(frame_duration));
  _time = image::Time(
    0, frame_duration ? (time_duration / frame_duration - 1) : 0,
    image::time::Speed(time_scale, frame_duration));
  DJV_DBG_PRN("time = " << _time);
  
  *info = _info;
  *time = _time;
  
  // Temporary buffers.
  
  _image.set(_info);
  
  err = ::NewGWorldFromPtr(&_gworld, k32RGBAPixelFormat, &rect, 0, 0, 0,
    (char *)_image.data(0, 0), _info.size.x * 4);
  if (err != noErr)
    throw Error(::name, str::Format(error_gworld).arg(err));
  ::SetMovieGWorld(_movie, _gworld, 0);
}

void Load::_close()
{
  DJV_DBG_OBJ2(_DBG, "Load::_close");
  
  if (_movie) { ::DisposeMovie(_movie); _movie = 0; }
  if (_gworld) { ::DisposeGWorld(_gworld); _gworld = 0; }
}

const image::Image * Load::get(int frame, int, int proxy) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);

  if (-1 == frame) frame = 0;

  // Read.
  
  const ::TimeValue time = frame * _time.speed.duration;
  DJV_DBG_PRN("time = " << int(time));
  ::SetMovieTimeValue(_movie, time);
  ::MoviesTask(_movie, 0);
  ::UpdateMovie(_movie);
  
  if (proxy)
  {
    image::Info info = _info;
    info.size = image::data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _tmp.set(info);
    image::data::proxy_scale(_image, &_tmp, proxy);
  }
  
  DJV_DBG_PRN("image = " << _image);
  return proxy ? &_tmp : &_image;
}

