/*------------------------------------------------------------------------------
 plugin/djv_quicktime_unix/load.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_quicktime_unix/load.cxx"

#include "plugin.h"

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_load() { return new Load; }
}

Load::Load() : _frame(0), _f(0) {}
Load::~Load() { _close(); }

void Load::open(const File & in, image::Info * info, image::Tag * tag,
  image::Time * time) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _close();
  
  // Open.
  
  _file = in;
  _f = lqt_open_read(_file.get().c_str());
  if (! _f)
    throw Error(name(), image::io::error_open);
  
  if (! quicktime_has_video(_f))
    throw Error(name(), error_video);

  if (! quicktime_supported_video(_f, 0))
    throw Error(name(), image::io::error_unsupported);

  // Information.
  
  _info.size = Vec2i(quicktime_video_width(_f, 0),
    quicktime_video_height(_f, 0));
  
  static int cmodel_list [] = {
    BC_RGB888,
    BC_BGR888,
    BC_RGB161616,
    BC_RGBA8888,
    BC_RGBA16161616,
    LQT_COLORMODEL_NONE };
  const int cmodel = lqt_get_best_colormodel(_f, 0, cmodel_list);
  switch (cmodel)
  {
    case BC_BGR888: _info.color = image::data::BGR;
    case BC_RGB888: _info.pixel = image::pixel::RGB_U8; break;
    case BC_RGB161616: _info.pixel = image::pixel::RGB_U16; break;
    case BC_RGBA8888: _info.pixel = image::pixel::RGBA_U8; break;
    case BC_RGBA16161616: _info.pixel = image::pixel::RGBA_U16; break;
    default: throw Error(name(), image::io::error_unsupported);
  }
  lqt_set_cmodel(_f, 0, cmodel);
  
  int time_constant = 0;
  const int time_scale = lqt_video_time_scale(_f, 0);
  const int frame_duration = lqt_frame_duration(_f, 0, &time_constant);
  _time = image::Time(
    0, frame_duration ? (lqt_video_duration(_f, 0) / frame_duration - 1) : 0,
    image::time::Speed(time_scale, frame_duration));
  DJV_DBG_PRN("time = " << _time);
  DJV_DBG_PRN("time constant = " << time_constant);
  
  const List<Str> & label_tag = image::tag::label();
  char * _tag = lqt_get_artist(_f);
  if (_tag) (*tag)[label_tag[image::tag::CREATOR]] = _tag;
  _tag = lqt_get_comment(_f);
  if (_tag) (*tag)[label_tag[image::tag::DESCRIPTION]] = _tag;
  _tag = quicktime_get_copyright(_f);
  if (_tag) (*tag)[label_tag[image::tag::COPYRIGHT]] = _tag;
  
  *info = _info;
  *time = _time;
}

void Load::_close()
{
  DJV_DBG_OBJ2(_DBG, "Load::_close");
  
  if (_f) { quicktime_close(_f); _f = 0; }
}

const image::Image * Load::get(int frame, int, int proxy) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  if (-1 == frame) frame = 0;

  image::Data * data = proxy ? &_tmp : &_image;
  data->set(_info);
  
  struct Rows
  {
    Rows(image::Data * in) : p(0)
    {
      p = new unsigned char * [in->h() + 1];
      for (int y = 0; y < in->h(); ++y)
        p[y] = reinterpret_cast<image::pixel::U8_T *>(
          in->data(0, in->h() - 1 - y));
    }
    
    ~Rows() { if (p) delete [] p; }
    
    unsigned char ** p;

  } rows(data);

  // XXX
  
  if (frame != _frame + 1)
  {
    const uint64_t time = frame * _time.speed.duration;
    DJV_DBG_PRN("time = " << int(time));
    lqt_seek_video(_f, 0, time);
  }
  _frame = frame;
  
  lqt_decode_video(_f, rows.p, 0);
  
  if (proxy)
  {
    image::Info info = _info;
    info.size = image::data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _image.set(info);
    image::data::proxy_scale(_tmp, &_image, proxy);
  }
  
  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

