/*------------------------------------------------------------------------------
 plugin/djv_tiff/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_tiff/save.cxx"

#include "plugin.h"

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save()
{
  tiff_init();
  return new Save;
}
}

Save::Save() : _f(0) {}
Save::~Save() { _close(); }

plugin::Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* tiff compress (value) - Compression: %1. Default: %2.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_compress(), ", ").
    arg(str::serialize(_options.compress, label_compress()));
}

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(data, &_options.compress, label_compress());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(_options.compress, label_compress());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const image::Info & info, const image::Tag &,
  const image::Time &) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  _info = image::Info();
  _info.size = info.size;
  _info.mirror_y = true;
  image::pixel::TYPE type = image::pixel::type(info.pixel);
  switch (type)
  {
    case image::pixel::U10: type = image::pixel::U16; break;
    case image::pixel::F16: type = image::pixel::F32; break;
    default: break;
  }
  _info.pixel = image::pixel::pixel(image::pixel::format(info.pixel), type);
  DJV_DBG_PRN("info = " << _info);
  
  _image.set(_info);
}

void Save::_open(const File & in, const image::Tag & tag) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  _close();
  
  // Open.
  
  _f = TIFFOpen(in.get().c_str(), "w");
  if (! _f)
    throw Error(name(), image::io::error_open);
  
  // Header.
  
  uint16 photometric = 0, samples = 0, sample_depth = 0, sample_format = 0;
  uint16 extra_samples [] = { EXTRASAMPLE_ASSOCALPHA };
  uint16 extra_samples_size = 0;
  uint16 compress = 0;
  
  switch (image::pixel::format(_image.pixel()))
  {
    case image::pixel::L:
      photometric = PHOTOMETRIC_MINISBLACK;
      samples = 1;
      break;
    case image::pixel::LA:
      photometric = PHOTOMETRIC_MINISBLACK;
      samples = 2;
      extra_samples_size = 1;
      break;
    case image::pixel::RGB:
      photometric = PHOTOMETRIC_RGB;
      samples = 3;
      break;
    case image::pixel::RGBA:
      photometric = PHOTOMETRIC_RGB;
      samples = 4;
      extra_samples_size = 1;
      break;
  }
  
  switch (image::pixel::type(_image.pixel()))
  {
    case image::pixel::U8:
      sample_depth = 8;
      sample_format = SAMPLEFORMAT_UINT;
      break;
    case image::pixel::U16:
      sample_depth = 16;
      sample_format = SAMPLEFORMAT_UINT;
      break;
    case image::pixel::F16:
    case image::pixel::F32:
      sample_depth = 32;
      sample_format = SAMPLEFORMAT_IEEEFP;
      break;
    default: break;
  }
  
  switch (_options.compress)
  {
    case COMPRESS_NONE: compress = COMPRESSION_NONE; break;
    case COMPRESS_RLE: compress = COMPRESSION_PACKBITS; break;
    case COMPRESS_LZW: compress = COMPRESSION_LZW; break;
  }
  
  TIFFSetField(_f, TIFFTAG_IMAGEWIDTH, _image.w());
  TIFFSetField(_f, TIFFTAG_IMAGELENGTH, _image.h());
  TIFFSetField(_f, TIFFTAG_PHOTOMETRIC, photometric);
  TIFFSetField(_f, TIFFTAG_SAMPLESPERPIXEL, samples);
  TIFFSetField(_f, TIFFTAG_BITSPERSAMPLE, sample_depth);
  TIFFSetField(_f, TIFFTAG_SAMPLEFORMAT, sample_format);
  TIFFSetField(_f, TIFFTAG_EXTRASAMPLES, extra_samples_size, extra_samples);
  TIFFSetField(_f, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
  TIFFSetField(_f, TIFFTAG_COMPRESSION, compress);
  TIFFSetField(_f, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  
  // Image tags.
  
  const List<Str> & label_tag = image::tag::label();
  Str _tag = tag[label_tag[image::tag::CREATOR]];
  if (_tag.size()) TIFFSetField(_f, TIFFTAG_ARTIST, _tag.c_str());
  _tag = tag[label_tag[image::tag::COPYRIGHT]];
  if (_tag.size()) TIFFSetField(_f, TIFFTAG_COPYRIGHT, _tag.c_str());
  _tag = tag[label_tag[image::tag::TIME]];
  if (_tag.size()) TIFFSetField(_f, TIFFTAG_DATETIME, _tag.c_str());
  _tag = tag[label_tag[image::tag::DESCRIPTION]];
  if (_tag.size()) TIFFSetField(_f, TIFFTAG_IMAGEDESCRIPTION, _tag.c_str());
}

void Save::_close()
{
  if (_f) { TIFFClose(_f); _f = 0; }
}

void Save::set(const image::Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);

  // Open.
  
  _open(_file.get(frame), in.tag);
  
  // Write.
  
  const image::Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    image::gl::data::copy(in, &_image);
    p = &_image;
  }
  
  const int h = p->h();
  for (int y = 0; y < h; ++y)
    if (TIFFWriteScanline(_f, (tdata_t *)p->data(0, y), y) == -1)
      throw Error(name(), image::io::error_write);
  _close();
}

